#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import itertools

from matplotlib.patches import BoxStyle

from ads.dataset.label_encoder import DataFrameLabelEncoder
from sklearn.feature_selection import SelectKBest
from sklearn.feature_selection import chi2
from sklearn.preprocessing import MinMaxScaler

import matplotlib.font_manager
from ads.common import utils, logger
from ads.dataset.helper import (
    fix_column_names,
    convert_columns,
    get_feature_type,
    convert_to_html,
)
from ads.dataset.plot import Plotting
from ads.dataset.progress import DummyProgressBar
from ads.dataset.timeseries import Timeseries
from ads.type_discovery.type_discovery_driver import TypeDiscoveryDriver
from ads.type_discovery.typed_feature import (
    DateTimeTypedFeature,
    ContinuousTypedFeature,
    GISTypedFeature,
    ConstantTypedFeature,
    CreditCardTypedFeature,
    ZipcodeTypedFeature,
    PhoneNumberTypedFeature,
    OrdinalTypedFeature,
    CategoricalTypedFeature,
    DocumentTypedFeature,
    AddressTypedFeature,
)
from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)


class PandasDataset(object):
    """
    This class provides APIs that can work on a sampled dataset.
    """

    @runtime_dependency(module="geopandas", install_from=OptionalDependency.GEO)
    def __init__(
        self,
        sampled_df,
        type_discovery=True,
        types={},
        metadata=None,
        progress=DummyProgressBar(),
    ):
        self.client = None
        self.sampled_df = fix_column_names(sampled_df)
        self.correlation = None
        self.feature_dist_html_dict = {}
        self.feature_types = metadata if metadata is not None else {}
        self.world = geopandas.read_file(
            geopandas.datasets.get_path("naturalearth_lowres")
        )

        self.numeric_columns = self.sampled_df.select_dtypes(
            utils.numeric_pandas_dtypes()
        ).columns.values

        # run type discovery
        if len(self.feature_types) == 0:
            if len(types) != 0:
                # update feature types as it is for the types given by user
                self._update_feature_types(types.keys())
            if type_discovery:
                try:
                    #
                    # perform type-discovery
                    #
                    progress.update("Running data type discovery")
                    tdd = TypeDiscoveryDriver()
                    self.feature_types = {
                        col_name: tdd.discover(col_name, col_vals)
                        for col_name, col_vals in self.sampled_df.items()
                        if col_name not in types
                    }
                except Exception as e:
                    print(
                        f"An error occured while performing typed discovery on this dataset. Try running again with "
                        f"`type_discovery=False`"
                    )
                    raise e

        # convert dataframe columns to the data types discovered.
        self.sampled_df = convert_columns(self.sampled_df, self.feature_types)

        # update feature types for rest of the columns - no type discovery, does not include user overrides
        self._update_feature_types(
            set(self.sampled_df.columns.values) - set(self.feature_types.keys())
        )
        self.sampled_df = self.sampled_df.reset_index(drop=True)

    def _find_feature_subset(self, df, target_name, include_n_features=32):

        if len(df.columns) <= include_n_features:
            return self.sampled_df
        else:
            try:
                y = df[target_name]
                X = df.drop(columns=[target_name])
                X_columns = X.columns

                X = X.fillna(X.mode().iloc[0])
                X = DataFrameLabelEncoder().fit_transform(X)
                X = MinMaxScaler().fit_transform(X)

                from sklearn.impute import SimpleImputer

                imp_most_freq = SimpleImputer(strategy="most_frequent")
                X = imp_most_freq.fit_transform(X)

                est = SelectKBest(score_func=chi2, k=include_n_features)
                est.fit_transform(X, y)

                subset_features = [self.target.name] + list(
                    itertools.compress(X_columns, est.get_support())
                )

                return self.sampled_df.filter(subset_features, axis=1)

            except Exception as e:
                print("_find_feature_subset failed: ", str(e))
                return pd.DataFrame()

    def _update_multiple_outputs(self, out, msg):
        if isinstance(out, (list, tuple)):
            for o in out:
                o.value = msg
        else:
            self._update_multiple_outputs([out], msg)

    def _calculate_dataset_statistics(self, is_wide_dataset, out):
        #
        # first the missing values for non-wide datasets
        #
        df_missing = pd.DataFrame()
        df_skew = pd.DataFrame()

        if is_wide_dataset:
            df_missing = pd.DataFrame()  # empty dataframe when not calculating
            df_skew = pd.DataFrame()  # empty dataframe when not calculating
        else:
            #
            # count missing values
            #
            d = {column_name: np.nan for column_name in self.df.columns}  # default
            for column_name in self.df.columns:
                self._update_multiple_outputs(
                    out, f"calculating missing values (<code>{column_name}</code>)..."
                )
                d[column_name] = self.df[column_name].isna().sum()

            df_missing = pd.DataFrame.from_dict(d, orient="index", columns=["missing"])

            #
            # calculate skew
            #
            d = {column_name: np.nan for column_name in self.df.columns}  # default
            for column_name in self.numeric_columns:
                self._update_multiple_outputs(
                    out, f"calculating skew (<code>{column_name}</code>)..."
                )
                if len(self.df[column_name].dropna()) > 0:
                    d[column_name] = self.df[column_name].dropna().skew()
                else:
                    d[column_name] = np.nan
            df_skew = pd.DataFrame.from_dict(d, orient="index", columns=["skew"])
            self._update_multiple_outputs(out, "calculating dataset statistics...")
            for col in self.df.columns:
                if df_missing["missing"][col] == len(self.df[col]):
                    self.df[col] = self.df[col].astype("object")

        #
        # compute descriptive dataset statistics
        df_stats = self.df.describe(include="all").round(2)

        self._update_multiple_outputs(out, "transpose dataset statistics...")
        df_stats = df_stats.transpose()

        self._update_multiple_outputs(out, "finalizing dataset statistics...")
        df_stats = df_stats.fillna("")

        #
        # we join the stats with missing df if we computed that
        #
        self._update_multiple_outputs(
            out, "Assembling statistics into single result..."
        )
        if not df_missing.empty:
            df_stats = df_stats.join(df_missing).fillna("")
        if not df_skew.empty:
            df_stats = df_stats.join(df_skew).fillna("")

        return df_stats

    def _generate_features_html(
        self, is_wide_dataset, n_features, df_stats, visualizations_follow
    ):

        html = utils.get_bootstrap_styles()

        if is_wide_dataset:
            html += """<p>The dataset has too many columns ({:,}) to
                efficiently show feature visualizations, instead only showing table of
                statistics</p>""".format(
                n_features
            )

        html += "<p><b>&#x2022; Note</b> these are computed on the entire dataset.</p>"
        html += "<hr>"
        html += (
            df_stats.replace(np.nan, "")
            .style.set_table_styles(utils.get_dataframe_styles(max_width=125))
            .set_table_attributes("class=table")
            .format(
                lambda x: "{:.8g}".format(x)
                if ("float" in str(type(x))) or ("int" in str(type(x)))
                else x
            )
            .render()
        )

        if visualizations_follow:
            html += "<br><hr><h2>Feature Visualizations...</h2>"

        return html

    def _generate_warnings_html(
        self, is_wide_dataset, n_rows, n_features, df_stats, out, accordion
    ):

        #
        # create the "Warnings" accordion section:
        #  - show high cardinal categoricals
        #  - show high missing values
        #  - large number of zeros (not for wide datasets - slow to compute)
        #
        #

        accum = []

        ignored_feature_types = ["document"]

        # more than 5% missing is a warning
        #

        if "missing" in df_stats.columns:
            df_missing = df_stats[["missing"]][df_stats["missing"] != ""]
            if not df_missing.empty:
                # ignore document types
                for column_name, missing in df_missing.iterrows():
                    if (
                        self.feature_types[column_name]["type"]
                        not in ignored_feature_types
                    ):
                        missing_values = float(missing[0])
                        missing_pct = 100 * missing_values / n_rows
                        if missing_pct >= 5:
                            accum.append(
                                {
                                    "label": "missing",
                                    "message": f"<code>{column_name}</code> has {missing_values} ({missing_pct:.1f}%) missing values. Consider remove the column or replace null values.",
                                }
                            )
                        if missing_pct == 100:
                            accum.append(
                                {
                                    "label": "missing",
                                    "message": f"<code>{column_name}</code> is excluded from correlation computation due to {missing_values} ({missing_pct:.1f}%) missing values.",
                                }
                            )

        # abs skew > 20 skewness warning issues
        #

        if "skew" in df_stats.columns:
            df_skew = df_stats[["skew"]][df_stats["skew"] != ""]
            if not df_skew.empty:
                # ignore document types
                for column_name, skew in df_skew.iterrows():
                    if (
                        self.feature_types[column_name]["type"]
                        not in ignored_feature_types
                    ):
                        skewness = float(skew[0])
                        if abs(skewness) >= 20:
                            accum.append(
                                {
                                    "label": "skew",
                                    "message": f"<code>{column_name}</code> has skew of {skewness:.3f}",
                                }
                            )

        # high cardinality (> 15 unique values) is a warning
        #

        if "unique" in df_stats.columns:
            for column_name, count in df_stats[["unique"]][
                df_stats["unique"] != ""
            ].iterrows():
                # ignore document types
                if self.feature_types[column_name]["type"] not in ignored_feature_types:
                    out.value = (
                        f"Analyzing cadinalities (<code>{column_name}</code>)..."
                    )
                    unique = self.df[column_name].unique().shape[0]
                    if unique == n_rows:
                        accum.append(
                            {
                                "label": "high-cardinality",
                                "message": f"<code>{column_name}</code> has a high cardinality: every value is distinct",
                            }
                        )
                    elif unique > 15:
                        accum.append(
                            {
                                "label": "high-cardinality",
                                "message": f"<code>{column_name}</code> has a high cardinality: {unique} distinct values",
                            }
                        )

        if not is_wide_dataset:
            # more than 10% zeros is a warning
            if "min" in df_stats.columns:
                for column_name, count in df_stats[["min"]][
                    df_stats["min"] != ""
                ].iterrows():
                    if (
                        self.feature_types[column_name]["type"]
                        not in ignored_feature_types
                    ):
                        out.value = f"Analyzing zeros (<code>{column_name}</code>)..."
                        # we've filtered on only columns that have a min value of 0 for speed
                        zeros = self.df[self.df[column_name] == 0].shape[0]
                        zeros_pct = 100 * zeros / n_rows
                        if zeros_pct >= 10:
                            accum.append(
                                {
                                    "label": "zeros",
                                    "message": f"<code>{column_name}</code> has {zeros} ({zeros_pct:.2f}%) zeros)",
                                }
                            )

        #
        # collect the warnings into an HTML presentation
        #

        out.value = "Assembling results..."

        if accum:
            html = utils.get_bootstrap_styles()

            html += utils.highlight_text(f"{len(accum)} WARNING(S) found")

            accordion.set_title(3, f"Warnings ({len(accum)})")

            html += "<hr>"

            tr_rows = [
                f"""
                <tr style="border-top:0">
                       <td>
                        {feature_warning['message']}
                       <td>
                        <span class="label label-{feature_warning['label']}">{feature_warning['label']}</span>
                       </td>
                </tr>
            """.strip()
                for feature_warning in accum
            ]

            html += """
                <table style="width: 100%; max-width: 100%;">
                <tbody>
                {}
                </tbody>
                </table>""".format(
                "\n".join(tr_rows)
            )

            return html

        else:
            return "<h3>No Feature warnings found</h3>"

    def summary(self, feature_name=None):
        """
        Display list of features & their datatypes.
        Shows the column name and the feature's meta_data if given a specific feature name.

        Parameters
        ----------
        date_col: str
            The name of the feature

        Returns
        -------
        dict
            a dictionary that contains requested information
        """

        feature_n_datatype = {}
        list_of_dfs = []

        if feature_name is None:
            feature_n_datatype = {
                col: self.feature_types[col].type
                + "/"
                + self.feature_types[col].low_level_type
                for col in self.sampled_df.columns
            }
            df = pd.DataFrame(
                feature_n_datatype.items(), columns=["Feature", "Datatype"]
            )
        else:
            if isinstance(feature_name, (list, tuple, pd.core.indexes.base.Index)):
                feature_names = list(feature_name)
            else:
                feature_names = [feature_name]

            for col in feature_names:
                if col in self.sampled_df.columns:
                    feature_n_datatype[col] = {
                        k: v
                        for k, v in self.feature_types[col].meta_data.items()
                        if k not in ["internal", "feature_name"]
                    }
                    new_dict = utils.flatten(feature_n_datatype[col])
                    tmp_df = pd.DataFrame.from_dict(
                        new_dict,
                        orient="index",
                        columns=[feature_names[feature_names.index(str(col))]],
                    )
                    list_of_dfs.append(tmp_df)
                else:
                    feature_n_datatype[col] = None

            df = pd.concat(list_of_dfs, axis=1).transpose().fillna("-")

            # get all the unique types from df
            new_list_dfs = []
            for t in df.type.unique():
                new_list_dfs.append(df[df["type"] == t])
            df = pd.concat(new_list_dfs)

            # reorder columns in df

            # get a list of columns
            cols = list(df)
            # move the column to head of list using index, pop and insert
            cols.insert(0, cols.pop(cols.index("low_level_type")))
            cols.insert(0, cols.pop(cols.index("type")))
            df = df.loc[:, cols]

        return df

    def timeseries(self, date_col):
        """
        Supports any plotting operations where x=datetime.

        Parameters
        ----------
        date_col: str
            The name of the feature to plot

        Returns
        -------
        func
            a plotting object that contains a date column and dataframe
        """

        if date_col in self.feature_types and isinstance(
            self.feature_types[date_col], DateTimeTypedFeature
        ):
            return Timeseries(date_col, self.sampled_df)
        else:
            raise ValueError("Not a date time column.")

    def plot(
        self, x, y=None, plot_type="infer", yscale=None, verbose=True, sample_size=0
    ):
        """
        Supports plotting feature distribution, and relationship between features.

        Parameters
        ----------
        x: str
            The name of the feature to plot
        y: str, optional
            Name of the feature to plot against x
        plot_type: str, default: infer
            Override the inferred plot type for certain combinations of the data types of x and y.
            By default, the best plot type is inferred based on x and y data types.
            Valid values:

            - box_plot - discrete feature vs continuous feature. Draw a box plot to show
              distributions with respect to categories,
            - scatter - continuous feature vs continuous feature. Draw a scatter plot
              with possibility of several semantic groupings.

        yscale : str, optional
            One of {"linear", "log", "symlog", "logit"}.
            The y axis scale type to apply. Can be used when either x or y is an ordinal feature.
        verbose: bool, default True
            Displays Note/Tips if True
        """
        sample_size = int(sample_size)
        min_sample_size = 10000
        if sample_size == 0:
            sub_samp_size = len(self.sampled_df)
            sub_samp_df = self.sampled_df
        else:
            sub_samp_size = max(min(sample_size, len(self.sampled_df)), min_sample_size)
            sub_samp_df = self.sampled_df.sample(n=sub_samp_size)
        plot = Plotting(
            sub_samp_df, self.feature_types, x, y=y, plot_type=plot_type, yscale=yscale
        )
        if verbose:
            if len(self.df) != sub_samp_size:
                logger.info(f"Downsampling from dataset for graphing.")
        return plot

    @runtime_dependency(module="geopandas", install_from=OptionalDependency.GEO)
    def plot_gis_scatter(self, lon="longitude", lat="latitude", ax=None):
        """
        Supports plotting Choropleth maps

        Parameters
        ----------
        df: pandas dataframe
            The dataframe to plot
        x: str
            The name of the feature to plot, usually the longitude
        y: str
            THe name of the feature to plot, usually the latitude
        """
        if lon in self.sampled_df.columns and lat in self.sampled_df.columns:
            if ax is None:
                fig, ax = plt.subplots(1, figsize=(10, 10))
            gdf = geopandas.GeoDataFrame(
                self.sampled_df,
                geometry=geopandas.points_from_xy(
                    self.sampled_df[lon], self.sampled_df[lat]
                ),
            )
            world = geopandas.read_file(
                geopandas.datasets.get_path("naturalearth_lowres")
            )
            ax1 = world.plot(ax=ax, color="lightgrey", linewidth=0.5, edgecolor="white")
            gdf.plot(ax=ax1, color="blue", markersize=10)

        else:
            if len(self.sampled_df.columns) > 0:
                logger.info(
                    "The available latitude and longitude columns are: "
                    + ", ".join(self.sampled_df.columns)
                    + "."
                )
            else:
                logger.info("There are no latitude and longitude columns available.")

    """
    Internal methods
    """

    def _update_feature_types(self, columns):
        # Build feature types for columns which are not type discovered, by using the inferred type as it is
        for column in columns:
            self.feature_types[column] = get_feature_type(
                column, self.sampled_df[column]
            )

    @runtime_dependency(module="geopandas", install_from=OptionalDependency.GEO)
    def _visualize_feature_distribution(self, html_widget):
        """
        This function is called once per dataset to generate html for feature distribution plots.
        """
        if len(self.feature_dist_html_dict) > 0:
            return self.feature_dist_html_dict

        feature_dist_html = ""
        figsize = (6.5, 2)

        props = {
            "boxstyle": BoxStyle("Round", pad=0),
            "facecolor": "white",
            "linestyle": "solid",
            "linewidth": 0,
            "edgecolor": "white",
        }

        font = {"size": 10}

        matplotlib.rc("font", **font)

        red_square = dict(markerfacecolor="r", marker="s")

        blues = [
            "#AED6F1",
            "#85C1E9",
            "#5DADE2",
            "#3498DB",
            "#2E86C1",
            "#2874A6",
            "#1B4F72",
        ]
        bright_colors = ["red", "green", "blue", "yellow", "green"]

        for col_index, col in enumerate(self.sampled_df.columns):
            feature_metadata = self.feature_types[col].meta_data
            text = "\n".join(
                [
                    f"{col}\n",
                    "  - type: {} ({})".format(
                        self.feature_types[col].type,
                        self.feature_types[col].low_level_type,
                    ),
                    "  - missing_percentage: {:.1f}%".format(
                        self.feature_types[col].missing_percentage
                    ),
                ]
            )

            fig, ax = PandasDataset._init_fig_ax(figsize)

            if isinstance(self.feature_types[col], ContinuousTypedFeature):
                text += PandasDataset._format_stats(
                    self.feature_types[col].type, feature_metadata["stats"]
                )
                ax.text(
                    -1.1,
                    1,
                    text,
                    transform=ax.transAxes,
                    verticalalignment="top",
                    bbox=props,
                )
                self.sampled_df[col].plot(
                    kind="box",
                    vert=False,
                    flierprops=red_square,
                    ax=ax,
                    figsize=figsize,
                )

            elif isinstance(self.feature_types[col], DateTimeTypedFeature):
                text += PandasDataset._format_stats(
                    self.feature_types[col].type, feature_metadata["stats"]
                )
                ax.text(
                    -1.1,
                    1,
                    text,
                    transform=ax.transAxes,
                    verticalalignment="top",
                    bbox=props,
                )
                from matplotlib.dates import AutoDateFormatter, AutoDateLocator

                loc = AutoDateLocator(interval_multiples=False)
                ax.xaxis.set_major_locator(loc)
                ax.xaxis.set_major_formatter(AutoDateFormatter(loc))
                self.sampled_df[col].hist(
                    bins=50,
                    grid=False,
                    xrot=45,
                    ax=ax,
                    rwidth=0.95,
                    color=blues[-1],
                    figsize=figsize,
                )

            elif isinstance(self.feature_types[col], GISTypedFeature):
                text += PandasDataset._format_stats(
                    self.feature_types[col].type, feature_metadata["stats"]
                )
                ax.text(
                    -1.1,
                    1,
                    text,
                    transform=ax.transAxes,
                    verticalalignment="top",
                    bbox=props,
                )
                df = pd.DataFrame(
                    feature_metadata["internal"]["sample"], columns=["lat", "lon"]
                )
                gdf = geopandas.GeoDataFrame(
                    df, geometry=geopandas.points_from_xy(df["lon"], df["lat"])
                )
                self.world.plot(
                    ax=ax, color="lightgrey", linewidth=0.5, edgecolor="white"
                )
                gdf.plot(ax=ax, color="blue", markersize=10)
                ax.set_aspect("auto")

            elif (
                isinstance(self.feature_types[col], ConstantTypedFeature)
                and feature_metadata["missing_percentage"] < 100
            ):
                text += PandasDataset._format_stats(
                    self.feature_types[col].type, feature_metadata["stats"]
                )
                ax.text(
                    -1.1,
                    1,
                    text,
                    transform=ax.transAxes,
                    verticalalignment="top",
                    bbox=props,
                )
                pd.Series(
                    feature_metadata["internal"]["counts"],
                    name=col,
                    index=feature_metadata["internal"]["counts"].keys(),
                ).plot(kind="barh", ax=ax, width=0.95, figsize=figsize, color=["black"])

            elif isinstance(self.feature_types[col], CreditCardTypedFeature):
                text += PandasDataset._format_stats(
                    self.feature_types[col].type, feature_metadata["stats"]
                )
                ax.text(
                    -1.1,
                    1,
                    text,
                    transform=ax.transAxes,
                    verticalalignment="top",
                    bbox=props,
                )
                sorted_by_value = sorted(
                    feature_metadata["internal"]["counts"],
                    key=feature_metadata["internal"]["counts"].get,
                    reverse=True,
                )
                pd.Series(
                    feature_metadata["internal"]["counts"],
                    name=col,
                    index=sorted_by_value,
                ).plot(kind="bar", ax=ax, width=0.95, figsize=figsize, color=blues)

            elif isinstance(self.feature_types[col], ZipcodeTypedFeature):
                ax.text(
                    -1.1,
                    1,
                    text,
                    transform=ax.transAxes,
                    verticalalignment="top",
                    bbox=props,
                )
                pd.Series(
                    feature_metadata["internal"]["histogram"],
                    name=col,
                    index=feature_metadata["internal"]["histogram"].keys(),
                ).plot(kind="bar", ax=ax, figsize=figsize, color=blues)

            elif isinstance(self.feature_types[col], PhoneNumberTypedFeature):
                text += PandasDataset._format_stats(
                    self.feature_types[col].type, feature_metadata["stats"]
                )
                ax.text(
                    -1.1,
                    1,
                    text,
                    transform=ax.transAxes,
                    verticalalignment="top",
                    bbox=props,
                )
                pd.Series(
                    feature_metadata["internal"]["counts"],
                    name=col,
                    index=feature_metadata["internal"]["counts"].keys(),
                ).plot(kind="bar", ax=ax, figsize=figsize, color=blues)

            elif isinstance(self.feature_types[col], OrdinalTypedFeature):
                text += PandasDataset._format_stats(
                    self.feature_types[col].type, feature_metadata["stats"]
                )
                high_cardinality = feature_metadata["internal"]["high_cardinality"]
                very_high_cardinality = feature_metadata["internal"][
                    "very_high_cardinality"
                ]

                ax.text(
                    -1.1,
                    1,
                    text,
                    transform=ax.transAxes,
                    verticalalignment="top",
                    bbox=props,
                )

                if very_high_cardinality:
                    addrtext = "Samples:\n\n"
                    addrtext += ", ".join(
                        feature_metadata["internal"]["counts"]
                        .keys()
                        .astype(str)
                        .to_list()[:6]
                    )

                    ax.text(
                        0.05,
                        0.95,
                        addrtext,
                        transform=ax.transAxes,
                        fontsize=10,
                        verticalalignment="top",
                        bbox=props,
                    )
                    ax.axis("off")

                else:
                    if high_cardinality:
                        text += (
                            "\n\n  NOTE: plot has been capped\n  from %d to show only most\n  common top %d "
                            "categories"
                            % (
                                feature_metadata["internal"]["unique"],
                                len(feature_metadata["internal"]["counts"].keys()),
                            )
                        )

                    if feature_metadata["internal"]["unique"] < 24:
                        pd.Series(
                            feature_metadata["internal"]["counts"],
                            name=col,
                            index=feature_metadata["internal"]["counts"].keys(),
                        ).plot(
                            kind="bar",
                            ax=ax,
                            width=0.90,
                            color=blues[-1],
                            figsize=figsize,
                        )
                    else:
                        self.sampled_df[col].plot(
                            kind="hist",
                            grid=False,
                            rwidth=0.95,
                            ax=ax,
                            color=blues[-1],
                            figsize=figsize,
                        )

            elif isinstance(self.feature_types[col], CategoricalTypedFeature):
                text += PandasDataset._format_stats(
                    self.feature_types[col].type, feature_metadata["stats"]
                )

                high_cardinality = feature_metadata["internal"]["high_cardinality"]
                very_high_cardinality = feature_metadata["internal"][
                    "very_high_cardinality"
                ]

                ax.text(
                    -1.1,
                    1,
                    text,
                    transform=ax.transAxes,
                    verticalalignment="top",
                    bbox=props,
                )

                if very_high_cardinality:
                    # grab the first few examples as strings
                    addrtext = "Samples:\n\n"
                    addrtext += "\n".join(
                        [
                            utils.ellipsis_strings(x, 65)
                            for x in feature_metadata["internal"]["counts"]
                            .keys()
                            .astype(str)
                            .to_list()[:3]
                        ]
                    )
                    ax.text(
                        0.05,
                        0.95,
                        addrtext,
                        transform=ax.transAxes,
                        fontsize=10,
                        verticalalignment="top",
                        bbox=props,
                    )
                    ax.axis("off")

                else:
                    text += PandasDataset._format_stats(
                        self.feature_types[col].type, feature_metadata["stats"]
                    )
                    if high_cardinality:
                        text += (
                            "\n\n  NOTE: plot has been capped\n  to show only most\n  common top %d categories"
                            % (len(feature_metadata["internal"]["counts"].keys()))
                        )

                    if feature_metadata["internal"]["unique"] == 2:
                        #
                        # binary
                        #
                        count_series = pd.Series(
                            feature_metadata["internal"]["counts"],
                            name=col,
                            index=feature_metadata["internal"]["counts"].keys(),
                        ).astype(float)
                        ax1 = count_series.plot(
                            kind="barh",
                            ax=ax,
                            width=0.95,
                            figsize=figsize,
                            color=[blues[0], blues[-1]],
                        )
                        # x_labels = utils.ellipsis_strings(feature_metadata['internal']['counts'].keys().astype(str))
                        # ax1.set_xticklabels(x_labels)
                    else:
                        #
                        # multiclass, potentially high cardinality
                        #

                        ax1 = pd.Series(
                            feature_metadata["internal"]["counts"],
                            name=col,
                            index=feature_metadata["internal"]["counts"].keys(),
                        ).plot(
                            kind="bar", ax=ax, width=0.95, color=blues, figsize=figsize
                        )

                        x_labels = utils.ellipsis_strings(
                            feature_metadata["internal"]["counts"].keys()
                        )
                        ax1.set_xticklabels(x_labels)

            elif isinstance(
                self.feature_types[col], DocumentTypedFeature
            ) or isinstance(self.feature_types[col], AddressTypedFeature):
                text += PandasDataset._format_stats(
                    self.feature_types[col].type,
                    {k: v for k, v in feature_metadata["stats"].items()},
                )

                ax.text(
                    -1.1,
                    1,
                    text,
                    transform=ax.transAxes,
                    verticalalignment="top",
                    bbox=props,
                )
                if "word_frequencies" in feature_metadata["internal"]:
                    word_freqs = feature_metadata["internal"]["word_frequencies"]
                    stats = "\n".join(["  - word count: %d" % (len(word_freqs.keys()))])
                    text = text + "\n"
                    text += stats
                    try:
                        from wordcloud import WordCloud, STOPWORDS

                        wordcloud = WordCloud(
                            width=1000,
                            height=int(1000 * (figsize[1] / figsize[0])),
                            background_color="white",
                            stopwords=set(STOPWORDS),
                            max_words=50,
                            max_font_size=75,
                        ).fit_words(word_freqs)

                        plt.imshow(wordcloud, interpolation="bilinear")
                        plt.axis("off")
                    except ModuleNotFoundError as e:
                        utils._log_missing_module("wordcloud", OptionalDependency.TEXT)
                        logger.info(
                            "The text word cloud is not plotted due to missing dependency wordcloud."
                        )

            else:
                ax.text(
                    -1.1,
                    1,
                    text,
                    transform=ax.transAxes,
                    verticalalignment="top",
                    bbox=props,
                    weight="bold",
                )
                if feature_metadata["missing_percentage"] == 100:
                    addrtext = "NOTE: plot has been disabled as all values in this column are missing."
                else:
                    addrtext = "NOTE: plot has been disabled,\nfor features of unknown type\nno visualization is available"
                ax.text(
                    0.05,
                    0.95,
                    addrtext,
                    transform=ax.transAxes,
                    fontsize=14,
                    verticalalignment="top",
                    bbox=props,
                    weight="bold",
                )
                ax.axis("off")

            self.feature_dist_html_dict[col] = convert_to_html(plt)
            plt.close()

            html_widget.value += self.feature_dist_html_dict[col]

    @staticmethod
    def _init_fig_ax(figsize, dpi=288):
        fig, ax = plt.subplots(figsize=figsize, dpi=dpi)
        fig.set(facecolor="white")
        return fig, ax

    @staticmethod
    def _format_stats(feature_type_name, stats):
        text = "\n  - %s statistics:" % (feature_type_name)
        for k in list(stats.keys()):
            if "percentage" in k:
                text += "\n    - {}: {:.3f}%".format(k, stats[k])
            elif isinstance(stats[k], (int, np.int64)) or (
                isinstance(stats[k], float)
                and not np.isnan(stats[k])
                and stats[k] == int(stats[k])
            ):
                text += "\n    - {}: {:,}".format(k, int(stats[k]))
            elif isinstance(stats[k], bool):
                text += "\n    - {}: {s}".format(k, "yes" if stats[k] else "no")
            elif isinstance(stats[k], (float, np.float64)):
                text += "\n    - {}: {:.3f}".format(k, stats[k])
            else:
                text += "\n    - {}: {}".format(k, stats[k])

        return text



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import matplotlib.pyplot as plt
from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)


@runtime_dependency(module="optuna", install_from=OptionalDependency.OPTUNA)
def _get_edf_plot(study: "optuna.study.Study", fig_size: tuple):

    from optuna.visualization.matplotlib import plot_edf

    plot_edf(study)
    plt.show(block=False)



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/


from typing import Callable


def _get_original_func(func: callable) -> Callable:
    """The helper to retrieve the original function from the decorated one."""
    if func and hasattr(func, "__wrapped__"):
        return _get_original_func(func.__wrapped__)
    return func


class class_or_instance_method(classmethod):
    """Converts a function to be a class method or an instance depending on how it is called at runtime.

    To declare a class method, use this idiom:

        class C:
            @classmethod
            def f(obj, *args, **kwargs):
                ...
    It can be called either on the class (e.g. C.f()) or on an instance (e.g. C().f()).
    If it is called on the class C.f(), the first argument (obj) will be the class (aka. cls).
    If it is called on the instance C().f(), the first argument (obj) will be the instance (aka. self).

    """

    def __get__(self, instance, type_):
        delegate_get = super().__get__ if instance is None else self.__func__.__get__
        return delegate_get(instance, type_)



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from types import TracebackType
from typing import Optional
from typing import Tuple
from typing import Type


class _DeferredImportExceptionContextManager(object):  # pragma: no cover
    """Context manager to defer exceptions from imports.
    Catches :exc:`ImportError` and :exc:`SyntaxError`.
    If any exception is caught, this class raises an :exc:`ImportError` when being checked.
    """

    def __init__(self) -> None:
        self._deferred = None  # type: Optional[Tuple[Exception, str]]

    def __enter__(self) -> "_DeferredImportExceptionContextManager":
        """Enter the context manager.

        Returns:
            Itself.
        """
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[Exception]],
        exc_value: Optional[Exception],
        traceback: Optional[TracebackType],
    ) -> Optional[bool]:
        """Exit the context manager.
        Args:
            exc_type:
                Raised exception type. :obj:`None` if nothing is raised.
            exc_value:
                Raised exception object. :obj:`None` if nothing is raised.
            traceback:
                Associated traceback. :obj:`None` if nothing is raised.
        Returns:
            :obj:`None` if nothing is deferred, otherwise :obj:`True`.
            :obj:`True` will suppress any exceptions avoiding them from propagating.
        """
        if isinstance(exc_value, (ImportError, SyntaxError)):
            if isinstance(exc_value, ImportError):
                message = (
                    "Tried to import '{}' but failed. Make sure that the package is "
                    "installed correctly to use this feature. Actual error: {}."
                ).format(exc_value.name, exc_value)
            elif isinstance(exc_value, SyntaxError):
                message = (
                    "Tried to import a package but failed due to a syntax error in {}. "
                    "make sure that the Python version is correct to use this feature. Actual "
                    "error: {}."
                ).format(exc_value.filename, exc_value)
            else:
                assert False

            self._deferred = (exc_value, message)
            return True
        return None

    def is_successful(self) -> bool:
        """Return whether the context manager has caught any exceptions.

        Returns:
            :obj:`True` if no exceptions are caught, :obj:`False` otherwise.
        """
        return self._deferred is None

    def check(self) -> None:
        """Check whether the context manger has caught any exceptions.
        Raises:
            :exc:`ImportError`:
                If any exception was caught from the caught exception.
        """
        if self._deferred is not None:
            exc_value, message = self._deferred
            raise ImportError(message) from exc_value


def try_import() -> _DeferredImportExceptionContextManager:  # pragma: no cover
    """Create a context manager that can wrap imports of optional packages to defer exceptions.
    Returns:
        Deferred import context manager.
    """
    return _DeferredImportExceptionContextManager()



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import copy
import os
import shutil
import tempfile
from distutils import dir_util
from typing import Tuple, Dict
import shlex

import ads
from ads.common.auth import create_signer, AuthContext
from ads.common.oci_client import OCIClientFactory
from ads.jobs import (
    Job,
    DataScienceJobRun,
    DataScienceJob,
    ScriptRuntime,
    ContainerRuntime,
)
from ads.opctl import logger
from ads.opctl.backend.base import Backend
from ads.opctl.config.resolver import ConfigResolver
from ads.opctl.constants import DEFAULT_IMAGE_SCRIPT_DIR
from jinja2 import Environment, PackageLoader
from ads.opctl.distributed.common.cluster_config_helper import (
    ClusterConfigToJobSpecConverter,
)

REQUIRED_FIELDS = [
    "project_id",
    "compartment_id",
    "subnet_id",
    "block_storage_size_in_GBs",
    "shape_name",
]


class MLJobBackend(Backend):
    def __init__(self, config: Dict) -> None:
        """
        Initialize a MLJobBackend object given config dictionary.

        Parameters
        ----------
        config: dict
            dictionary of configurations
        """
        self.config = config
        self.oci_auth = create_signer(
            config["execution"].get("auth"),
            config["execution"].get("oci_config", None),
            config["execution"].get("oci_profile", None),
        )
        self.auth_type = config["execution"].get("auth")
        self.profile = config["execution"].get("oci_profile", None)
        self.client = OCIClientFactory(**self.oci_auth).data_science

    def apply(self) -> None:
        """
        Create Job and Job Run from YAML.
        """
        with AuthContext():
            ads.set_auth(auth=self.auth_type, profile=self.profile)
            job = Job.from_dict(self.config)
            job.create()
            job_run = job.run()
            print("JOB OCID:", job.id)
            print("JOB RUN OCID:", job_run.id)

    def run(self) -> None:
        """
        Create Job and Job Run from OCID or cli parameters.
        """
        # TODO Check that this still runs smoothly for distributed
        with AuthContext():
            ads.set_auth(auth=self.auth_type, profile=self.profile)
            if self.config["execution"].get("ocid", None):
                job_id = self.config["execution"]["ocid"]
                run_id = (
                    Job.from_datascience_job(self.config["execution"]["ocid"]).run().id
                )
            else:
                payload = self._create_payload()  # create job with infrastructure
                src_folder = self.config["execution"].get("source_folder")
                if self.config["execution"].get("conda_type") and self.config[
                    "execution"
                ].get("conda_slug"):
                    # add conda runtime
                    job_id, run_id = self._run_with_conda_pack(payload, src_folder)
                elif self.config["execution"].get("image"):
                    # add docker image runtime
                    job_id, run_id = self._run_with_image(payload)
                else:
                    raise ValueError(
                        "Either conda info or image name should be provided."
                    )
            print("JOB OCID:", job_id)
            print("JOB RUN OCID:", run_id)
            return {"job_id": job_id, "run_id": run_id}

    def init_operator(self):
        # TODO: check if folder is empty, check for force overwrite
        # TODO: check that command is being run from advanced-ds repo (important until ads released)

        operator_folder = self.config["execution"].get("operator_folder_path")
        os.makedirs(operator_folder, exist_ok=True)

        operator_folder_name = os.path.basename(os.path.normpath(operator_folder))
        docker_tag = f"{os.path.join(self.config['infrastructure'].get('docker_registry'), operator_folder_name)}:latest"

        self.config["execution"]["operator_folder_name"] = operator_folder_name
        self.config["execution"]["docker_tag"] = docker_tag

        operator_slug = self.config["execution"].get("operator_slug")
        self._jinja_write(operator_slug, operator_folder)

        # DONE
        print(
            "\nInitialization Successful.\n"
            f"All code should be written in main.py located at: {os.path.join(operator_folder, 'main.py')}\n"
            f"Additional libraries should be added to environment.yaml located at: {os.path.join(operator_folder, 'environment.yaml')}\n"
            "Any changes to main.py will require re-building the docker image, whereas changes to args in the"
            " runtime section of the yaml file do not. Write accordingly.\n"
            "Run this cluster with:\n"
            f"\tdocker build -t {docker_tag} -f {os.path.join(operator_folder, 'Dockerfile')} .\n"
            f"\tads opctl publish-image {docker_tag} \n"
            f"\tads opctl run -f {os.path.join(operator_folder, operator_slug + '.yaml')} \n"
        )
        return operator_folder

    def delete(self):
        """
        Delete Job or Job Run from OCID.
        """
        if self.config["execution"].get("id"):
            job_id = self.config["execution"]["id"]
            with AuthContext():
                ads.set_auth(auth=self.auth_type, profile=self.profile)
                Job.from_datascience_job(job_id).delete()
                print(f"Job {job_id} has been deleted.")
        elif self.config["execution"].get("run_id"):
            run_id = self.config["execution"]["run_id"]
            with AuthContext():
                ads.set_auth(auth=self.auth_type, profile=self.profile)
                DataScienceJobRun.from_ocid(run_id).delete()
                print(f"Job run {run_id} has been deleted.")

    def cancel(self):
        """
        Cancel Job Run from OCID.
        """
        run_id = self.config["execution"]["run_id"]
        with AuthContext():
            ads.set_auth(auth=self.auth_type, profile=self.profile)
            DataScienceJobRun.from_ocid(run_id).cancel()
            print(f"Job run {run_id} has been cancelled.")

    def watch(self):
        """
        Watch Job Run from OCID.
        """
        run_id = self.config["execution"]["run_id"]

        with AuthContext():
            ads.set_auth(auth=self.auth_type, profile=self.profile)
            run = DataScienceJobRun.from_ocid(run_id)
            run.watch()

    def _jinja_write(self, operator_slug, operator_folder):
        # TODO AH: fill in templates with relevant details
        env = Environment(
            loader=PackageLoader("ads", f"opctl/operators/{operator_slug}")
        )

        for setup_file in [
            "Dockerfile",
            "environment.yaml",
            "main.py",
            "run.py",
            "start_scheduler.sh",
            "start_worker.sh",
            "dask_cluster.yaml",
        ]:
            template = env.get_template(setup_file + ".jinja2")
            with open(os.path.join(operator_folder, setup_file), "w") as ff:
                ff.write(template.render(config=self.config))

    def _create_payload(self, infra=None, name=None) -> Job:
        if not infra:
            infra = self.config.get("infrastructure", {})
        # if any(k not in infra for k in REQUIRED_FIELDS):
        #    missing = [k for k in REQUIRED_FIELDS if k not in infra]
        #    raise ValueError(
        #        f"Following fields are missing but are required for OCI ML Jobs: {missing}. Please run `ads opctl configure`."
        #    )

        ml_job = DataScienceJob(spec=infra if "spec" not in infra else infra["spec"])

        log_group_id = infra.get("log_group_id")
        log_id = infra.get("log_id")

        if log_group_id:
            ml_job.with_log_group_id(log_group_id)
        if log_id:
            ml_job.with_log_id(log_id)
        if not name:
            try:
                name = infra.get("displayName") or self.config["execution"].get(
                    "job_name"
                )
            except:
                pass

        return Job(
            name=name,
            infrastructure=ml_job,
        )

    def _run_with_conda_pack(self, payload: Job, src_folder: str) -> Tuple[str, str]:
        payload.with_runtime(
            ScriptRuntime().with_environment_variable(
                **self.config["execution"]["env_vars"]
            )
        )
        if self.config["execution"].get("conda_type") == "service":
            payload.runtime.with_service_conda(self.config["execution"]["conda_slug"])
        else:
            payload.runtime.with_custom_conda(self.config["execution"]["conda_uri"])

        if ConfigResolver(self.config)._is_ads_operator():
            with tempfile.TemporaryDirectory() as td:
                os.makedirs(os.path.join(td, "operators"), exist_ok=True)
                dir_util.copy_tree(
                    src_folder,
                    os.path.join(td, "operators", os.path.basename(src_folder)),
                )
                curr_dir = os.path.dirname(os.path.abspath(__file__))
                shutil.copy(
                    os.path.join(curr_dir, "..", "operators", "run.py"),
                    os.path.join(td, "operators"),
                )
                payload.runtime.with_source(
                    os.path.join(td, "operators"), entrypoint="operators/run.py"
                )
                payload.runtime.set_spec(
                    "args", shlex.split(self.config["execution"]["command"] + " -r")
                )
                job = payload.create()
                job_id = job.id
                run_id = job.run().id
        else:
            with tempfile.TemporaryDirectory() as td:
                dir_util.copy_tree(
                    src_folder, os.path.join(td, os.path.basename(src_folder))
                )
                payload.runtime.with_source(
                    os.path.normpath(os.path.join(td, os.path.basename(src_folder))),
                    entrypoint=os.path.join(
                        os.path.basename(src_folder),
                        self.config["execution"]["entrypoint"],
                    ),
                )
                if self.config["execution"].get("command"):
                    payload.runtime.set_spec(
                        "args", shlex.split(self.config["execution"]["command"])
                    )
                job = payload.create()
                job_id = job.id
                run_id = job.run().id
        return job_id, run_id

    def _run_with_image(self, payload: Job) -> Tuple[str, str]:
        payload.with_runtime(
            ContainerRuntime().with_environment_variable(
                **self.config["execution"]["env_vars"]
            )
        )
        image = self.config["execution"]["image"]
        if ":" not in image:
            image += ":latest"
        payload.runtime.with_image(image)
        if os.path.basename(image) == image:
            logger.warn("Did you include registry in image name?")

        if ConfigResolver(self.config)._is_ads_operator():
            command = f"python {os.path.join(DEFAULT_IMAGE_SCRIPT_DIR, 'operators/run.py')} -r "
        else:
            command = ""
            # running a non-operator image
            if self.config["execution"].get("entrypoint"):
                payload.runtime.with_entrypoint(self.config["execution"]["entrypoint"])

        if self.config["execution"].get("command"):
            command += f"{self.config['execution']['command']}"
        if len(command) > 0:
            payload.runtime.with_cmd(",".join(shlex.split(command)))

        job = payload.create()
        job_id = job.id
        run_id = job.run().id
        return job_id, run_id


class MLJobDistributedBackend(MLJobBackend):
    DIAGNOSTIC_COMMAND = "python -m ads.opctl.diagnostics -t distributed"

    def __init__(self, config: Dict) -> None:
        """
        Initialize a MLJobDistributedBackend object given config dictionary.

        Parameters
        ----------
        config: dict
            dictionary of configurations
        """
        super().__init__(config=config)
        self.job = None

    def prepare_job_config(self, cluster_info):
        job_conf_helper = ClusterConfigToJobSpecConverter(cluster_info)
        jobdef_conf = job_conf_helper.job_def_info()
        infrastructure = cluster_info.infrastructure
        if jobdef_conf.get("name"):
            infrastructure["spec"]["displayName"] = jobdef_conf.get("name")
        job = self._create_payload(infrastructure["spec"])
        envVars = {}
        envVars.update(
            cluster_info.cluster.config.envVars
        )  # Add user provided environment variables
        envVars.update(
            jobdef_conf.get("envVars") or {}
        )  # Update with `OCI__` environment variables

        job.with_runtime(ContainerRuntime().with_environment_variable(**envVars))
        job.runtime.with_image(image=jobdef_conf["image"])
        self.job = job
        if os.path.basename(jobdef_conf["image"]) == jobdef_conf["image"]:
            logger.warning("Did you include registry in image name?")
        main_jobrun_conf = job_conf_helper.job_run_info("main")
        main_jobrun_conf["envVars"]["RANK"] = "0"
        main_jobrun_conf["name"] = main_jobrun_conf.get("name") or "main"

        worker_jobrun_conf = job_conf_helper.job_run_info("worker")
        worker_jobrun_conf_list = []
        if worker_jobrun_conf:
            for i in range(cluster_info.cluster.worker.replicas):
                conf = copy.deepcopy(worker_jobrun_conf)
                conf["envVars"]["RANK"] = str(i + 1)
                conf["name"] = (
                    conf.get("name", worker_jobrun_conf["envVars"]["OCI__MODE"])
                    + "_"
                    + str(i)
                )
                worker_jobrun_conf_list.append(conf)
        ps_jobrun_conf = job_conf_helper.job_run_info("ps")
        ps_jobrun_conf_list = []
        if ps_jobrun_conf:
            for i in range(cluster_info.cluster.ps.replicas):
                conf = copy.deepcopy(ps_jobrun_conf)
                conf["name"] = (
                    conf.get("name", worker_jobrun_conf["envVars"]["OCI__MODE"])
                    + "_"
                    + str(i)
                )
                ps_jobrun_conf_list.append(conf)

        worker_jobrun_conf_list.extend(ps_jobrun_conf_list)
        return main_jobrun_conf, worker_jobrun_conf_list

    @staticmethod
    def generate_worker_name(worker_jobrun_conf, i):
        return f"{worker_jobrun_conf['name']}-{i}"

    def run_diagnostics(self, cluster_info, dry_run=False, **kwargs):
        with AuthContext():
            ads.set_auth(auth=self.auth_type, profile=self.profile)
            main_jobrun_conf, worker_jobrun_conf_list = self.prepare_job_config(
                cluster_info=cluster_info
            )
            self.job.runtime.with_entrypoint(["/bin/bash", "--login", "-c"])
            self.job.runtime.with_cmd(MLJobDistributedBackend.DIAGNOSTIC_COMMAND)
            if dry_run:  # If dry run, print the job yaml on the console.
                print(
                    "-----------------------------Entering dryrun mode----------------------------------"
                )
                print(f"Creating Job with payload: \n{self.job}")
                print("+" * 200)

                print(f"Creating Main Job Run with following details:")
                print(f"Name: {main_jobrun_conf['name']}")
                print(f"Additional Environment Variables: ")
                main_env_Vars = main_jobrun_conf.get("envVars", {})
                for k in main_env_Vars:
                    print(f"\t{k}:{main_env_Vars[k]}")
                print("~" * 200)

                print(
                    "-----------------------------Ending dryrun mode----------------------------------"
                )
                return None
            else:
                job = self.job.create()

                # Start main job
                conf = dict(main_jobrun_conf)
                main_jobrun = job.run(
                    conf["name"],
                    env_var=conf["envVars"],
                    # freeform_tags={"distributed_training": "oracle-ads"},
                )
                self.job = job
                main_jobrun.watch()
                return job, main_jobrun

    def run(self, cluster_info, dry_run=False) -> None:
        """
        * Creates Job Definition  and starts main and worker jobruns from that job definition
        * The Job Definition will contain all the environment variables defined at the cluster/spec/config level, environment variables defined by the user at runtime/spec/env level and `OCI__` derived from the yaml specification
        * The Job Run will have overrides provided by the user under cluster/spec/{main|worker}/config section and `OCI__MODE`={MASTER|WORKER} depending on the run type
        """
        with AuthContext():
            ads.set_auth(auth=self.auth_type, profile=self.profile)
            main_jobrun_conf, worker_jobrun_conf_list = self.prepare_job_config(
                cluster_info=cluster_info
            )
            if dry_run:  # If dry run, print the job yaml on the console.
                print(
                    "-----------------------------Entering dryrun mode----------------------------------"
                )
                print(f"Creating Job with payload: \n{self.job}")
                print(
                    "++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++"
                )

                print(f"Creating Main Job Run with following details:")
                print(f"Name: {main_jobrun_conf['name']}")
                print(f"Additional Environment Variables: ")
                main_env_Vars = main_jobrun_conf.get("envVars", {})
                for k in main_env_Vars:
                    print(f"\t{k}:{main_env_Vars[k]}")
                print(
                    "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"
                )
                if cluster_info.cluster.worker:
                    print(f"Creating Job Runs with following details:")
                    for i in range(len(worker_jobrun_conf_list)):
                        worker_jobrun_conf = worker_jobrun_conf_list[i]
                        print("Name: " + worker_jobrun_conf.get("name"))
                        print("Additional Environment Variables: ")
                        worker_env_Vars = worker_jobrun_conf.get("envVars", {})
                        for k in worker_env_Vars:
                            print(f"\t{k}:{worker_env_Vars[k]}")

                print(
                    "-----------------------------Ending dryrun mode----------------------------------"
                )
                return None

            else:
                job = self.job.create()

                # Start main job
                conf = dict(main_jobrun_conf)
                main_jobrun = job.run(
                    conf["name"],
                    env_var=conf["envVars"],
                    # freeform_tags={"distributed_training": "oracle-ads"},
                )

                # Start worker job
                worker_jobruns = []
                if cluster_info.cluster.worker:
                    for i in range(len(worker_jobrun_conf_list)):
                        worker_jobrun_conf = worker_jobrun_conf_list[i]
                        conf = dict(worker_jobrun_conf)
                        jobrun = job.run(
                            worker_jobrun_conf.get("name"),
                            env_var=conf["envVars"],
                        )
                        worker_jobruns.append(jobrun)
                self.job = job
                return job, main_jobrun, worker_jobruns



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from typing import Dict
import ads
from ads.common.auth import create_signer, AuthContext
from ads.common.oci_client import OCIClientFactory
from ads.opctl.backend.base import Backend
from ads.pipeline import Pipeline, PipelineRun


class PipelineBackend(Backend):
    def __init__(self, config: Dict) -> None:
        """
        Initialize a MLPipeline object given config dictionary.

        Parameters
        ----------
        config: dict
            dictionary of configurations
        """
        self.config = config
        self.oci_auth = create_signer(
            config["execution"].get("auth"),
            config["execution"].get("oci_config", None),
            config["execution"].get("oci_profile", None),
        )
        self.auth_type = config["execution"].get("auth")
        self.profile = config["execution"].get("oci_profile", None)
        self.client = OCIClientFactory(**self.oci_auth).data_science

    def apply(self) -> None:
        """
        Create Pipeline and Pipeline Run from YAML.
        """
        with AuthContext():
            ads.set_auth(auth=self.auth_type, profile=self.profile)
            pipeline = Pipeline.from_dict(self.config)
            pipeline.create()
            pipeline_run = pipeline.run()
            print("PIPELINE OCID:", pipeline.id)
            print("PIPELINE RUN OCID:", pipeline_run.id)

    def run(self) -> None:
        """
        Create Pipeline and Pipeline Run from OCID.
        """
        pipeline_id = self.config["execution"]["ocid"]
        with AuthContext():
            ads.set_auth(auth=self.auth_type, profile=self.profile)
            pipeline = Pipeline.from_ocid(ocid=pipeline_id)
            pipeline_run = pipeline.run()
            print("PIPELINE RUN OCID:", pipeline_run.id)

    def delete(self) -> None:
        """
        Delete Pipeline or Pipeline Run from OCID.
        """
        if self.config["execution"].get("id"):
            pipeline_id = self.config["execution"]["id"]
            with AuthContext():
                ads.set_auth(auth=self.auth_type, profile=self.profile)
                Pipeline.from_ocid(pipeline_id).delete()
                print(f"Pipeline {pipeline_id} has been deleted.")
        elif self.config["execution"].get("run_id"):
            run_id = self.config["execution"]["run_id"]
            with AuthContext():
                ads.set_auth(auth=self.auth_type, profile=self.profile)
                PipelineRun.from_ocid(run_id).delete()
                print(f"Pipeline run {run_id} has been deleted.")

    def cancel(self) -> None:
        """
        Cancel Pipeline Run from OCID.
        """
        run_id = self.config["execution"]["run_id"]
        with AuthContext():
            ads.set_auth(auth=self.auth_type, profile=self.profile)
            PipelineRun.from_ocid(run_id).cancel()
            print(f"Pipeline run {run_id} has been cancelled.")

    def watch(self) -> None:
        """
        Watch Pipeline Run from OCID.
        """
        run_id = self.config["execution"]["run_id"]
        log_type = self.config["execution"]["log_type"]
        with AuthContext():
            ads.set_auth(auth=self.auth_type, profile=self.profile)
            PipelineRun.from_ocid(run_id).watch(log_type=log_type)



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import oci
import os


def download_certificates(auth, ca_file_name, cert_file_name, key_file_name):
    """ """
    client = oci.certificates.CertificatesClient(**auth)
    # Download Cert and Private Key
    print("Fetching certificate details")
    response = client.get_certificate_bundle(
        certificate_id=os.environ["OCI__CERTIFICATE_OCID"],
        certificate_bundle_type="CERTIFICATE_CONTENT_WITH_PRIVATE_KEY",
    )
    with open(cert_file_name, "w") as f:
        print(f"writing {cert_file_name}")
        f.write(response.data.certificate_pem)
    with open(key_file_name, "w") as f:
        print(f"writing {key_file_name}")
        f.write(response.data.private_key_pem)
    # Download CA Cert
    response = client.get_certificate_authority_bundle(
        certificate_authority_id=os.environ["OCI__CERTIFICATE_AUTHORITY_OCID"]
    )
    with open(ca_file_name, "w") as f:
        print(f"writing {ca_file_name}")
        f.write(response.data.cert_chain_pem)



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from ads.common import logger
from ads.hpo.ads_search_space import get_model2searchspace
from ads.hpo.distributions import *
from ads.hpo.utils import _is_arraylike, _make_indexable, _num_samples, _safe_indexing
from sklearn.exceptions import NotFittedError
from sklearn.pipeline import Pipeline


ADS_DISTRIBUTIONS = (
    UniformDistribution,
    LogUniformDistribution,
    DiscreteUniformDistribution,
    IntUniformDistribution,
    IntLogUniformDistribution,
    CategoricalDistribution,
)


def assert_model_is_supported(estimator):
    if estimator.__class__ not in get_model2searchspace().keys():
        raise NotImplementedError(
            "{} is not supported.".format(estimator.__class__.__name__)
        )


def assert_tuner_is_fitted(estimator, msg=None):
    if msg is None:
        msg = (
            "This %(name)s instance is not fitted yet. Call 'fit' with "
            "appropriate arguments before using this method."
        )

    if not hasattr(estimator, "tune"):
        raise TypeError("%s is not an estimator instance." % (estimator))

    attributes = ["_n_splits", "_sample_indices", "_scorer", "_study"]

    if not all([hasattr(estimator, attr) for attr in attributes]):
        raise NotFittedError(msg % {"name": type(estimator).__name__})


def assert_is_estimator(estimator):
    assert hasattr(estimator, "fit"), "Estimator must implement fit"


def validate_pipeline(model):
    if isinstance(model, list):
        assert all((isinstance(m, tuple) for m in model)), "Model is not a pipeline."
        return Pipeline(model)
    return model


def validate_search_space(params, param_distributions):
    assert isinstance(
        param_distributions, dict
    ), "Must pass a dictionary to <code>search_space</code>!"

    params_to_delete = []
    for param, distribution in param_distributions.items():
        if not param in params:
            logger.warning(
                f"Ignoring {param} as it is not a hyperparameter of the model."
            )
            params_to_delete.append(param)

        if (
            isinstance(distribution, float)
            or isinstance(distribution, int)
            or isinstance(distribution, str)
        ):
            param_distributions.update({param: CategoricalDistribution([distribution])})
        elif isinstance(distribution, list):
            param_distributions.update({param: CategoricalDistribution(distribution)})
        elif not isinstance(distribution, ADS_DISTRIBUTIONS):
            raise ValueError(
                "<code>search_space</code> only accept numbers, strings, list or distributions."
            )
    for param in params_to_delete:
        del param_distributions[param]


def assert_strategy_valid(param_distributions, new_strategy, old_strategy=None):
    if isinstance(new_strategy, str):
        assert new_strategy in [
            "perfunctory",
            "detailed",
        ], "Valid values of <code>strategy</code> are `perfunctory` and `detailed`."
    else:
        assert isinstance(
            new_strategy, dict
        ), "Valid <code>strategy</code> type are string and dictionary."
        if old_strategy:
            for name, distribution in new_strategy.items():
                if name in param_distributions.keys():
                    assert isinstance(
                        distribution, param_distributions[name].__class__
                    ), "Cannot change the distribution of existing params."
                    if isinstance(distribution, CategoricalDistribution):
                        assert set(param_distributions[name].choices) == set(
                            distribution.choices
                        ), "Does not support updating the list of values for categorical distributions."


def validate_fit_params(
    X,  # type: TwoDimArrayLikeType
    fit_params,  # type: Dict
    indices,  # type: OneDimArrayLikeType
):
    # type: (...) -> Dict

    fit_params_validated = {}
    for key, value in fit_params.items():

        # NOTE Original implementation:
        # https://github.com/scikit-learn/scikit-learn/blob/ \
        # 2467e1b84aeb493a22533fa15ff92e0d7c05ed1c/sklearn/utils/validation.py#L1324-L1328
        # Scikit-learn does not accept non-iterable inputs.
        # This line is for keeping backward compatibility.
        # (See: https://github.com/scikit-learn/scikit-learn/issues/15805)
        if not _is_arraylike(value) or _num_samples(value) != _num_samples(X):
            fit_params_validated[key] = value

        else:
            fit_params_validated[key] = _make_indexable(value)
            fit_params_validated[key] = _safe_indexing(
                fit_params_validated[key], indices
            )
    return fit_params_validated


def validate_params_for_plot(params, param_distributions):
    params = [] if params is None else params
    if not set(params).issubset(set(param_distributions.keys())):
        raise ValueError("Not all the params are in the search space.")



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from __future__ import print_function, absolute_import

import random
from collections import defaultdict
from math import pi
import pandas as pd
import numpy as np

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from matplotlib import colors as mcolors
from ads.dataset.helper import _log_yscale_not_set

from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)
from ads.common.utils import _log_plot_high_cardinality_warning, MAX_DISPLAY_VALUES
from ads.type_discovery.latlon_detector import LatLonDetector
from ads.type_discovery.typed_feature import (
    ContinuousTypedFeature,
    DateTimeTypedFeature,
    ConstantTypedFeature,
    DiscreteTypedFeature,
    CreditCardTypedFeature,
    ZipcodeTypedFeature,
    OrdinalTypedFeature,
    CategoricalTypedFeature,
    GISTypedFeature,
)

from ads.dataset import logger


class Plotting:
    def __init__(self, df, feature_types, x, y=None, plot_type="infer", yscale=None):
        self.df = df
        self.feature_types = feature_types
        self.x = x
        self.y = y
        self.x_type = self.feature_types[self.x]
        self.y_type = self.feature_types[self.y] if self.y is not None else None
        self.plot_type = plot_type
        self.yscale = yscale

    def __repr__(self):
        choices = self._get_plot_method()  # add (plot_type='{0}') in plot method"
        if len(choices) > 1:
            logger.info(f"Recommended plot type is {choices[0][1].__name__}.")

            logger.info(
                "Available plot types are ",
                ", ".join([x[1].__name__ for x in choices]),
                ".",
            )
        self.show_in_notebook()
        return ""

    def select_best_plot(self):
        """
        Returns the best plot for a given dataset
        """

        #
        # auto logic
        #

        choices = self._get_plot_method()

        if len(choices) > 1:
            logger.info(
                "select_best_plot (%s, %s) called, possible plot types are %s"
                % (
                    self.x_type.meta_data["type"],
                    self.y_type.meta_data["type"] if self.y_type is not None else "",
                    ", ".join([x[0].__name__ for x in choices]),
                )
            )

        if self.plot_type != "infer":
            for choice in choices:
                if choice[1].__name__.lower().startswith(self.plot_type.lower()):
                    return choice
            logger.info("invalid plot_type: {}".format(self.plot_type))
            raise ValueError(
                "plot_type: '%s' invalid, use one of: %s"
                % (self.plot_type, ", ".join([x[0].__name__ for x in choices]))
            )

        return choices[0]

    def show_in_notebook(self, **kwargs):

        """
        Visualizes the dataset by plotting the distribution of a feature or relationship between two features.

        Parameters
        ----------
        figsize: tuple
                defines the size of the fig
        -------
        """
        plotlib_type, plot_method, plot_kwargs = self.select_best_plot()
        plotlib_type(plot_method, **plot_kwargs, **kwargs)

    @staticmethod
    def _add_identity(ax, *line_args, **line_kwargs):
        ax.plot(ax.get_xlim(), ax.get_ylim(), ls="--", c="-.3", transform=ax.transAxes)

    @staticmethod
    def _build_plot_key(x_type, y_type=None):
        if y_type is None:
            return x_type.__name__
        return x_type.__name__ + "," + y_type.__name__

    @staticmethod
    @runtime_dependency(module="scipy", install_from=OptionalDependency.VIZ)
    def _gaussian_heatmap(x, y, data, s=10, edgecolor="white", cmap=plt.cm.jet):
        """
        Generate a scatter plot and assign a color to each data point based on the local density (gaussian kernel) of
        points.

        Parameters
        ----------
        x: str
            name of the feature
        y: str
            name of the feature
        data: object
            The dataframe that contains x and y
        s: int
            area of each marker
        edgecolor: str
            edge color of each point. string value, e.g. 'blue'
        cmap: object
            color map for the heatmap

        Returns
        -------
        ax.scatter() object: a scatter plot with colored density

        Raises
        ------
        ValueError
            When the columns are identical or the columns are highly correlated.
        """
        try:
            _x = np.array(data[x])
            _y = np.array(data[y])
            xy = np.vstack([_x, _y])
            z = scipy.stats.gaussian_kde(xy)(xy)
            sc = plt.scatter(_x, _y, c=z, s=s, edgecolor=edgecolor, cmap=cmap)
            plt.xlabel(x)
            plt.ylabel(y)
            return plt.colorbar(sc)
        except:
            return plt.scatter(_x, _y)

    @staticmethod
    @runtime_dependency(module="seaborn", install_from=OptionalDependency.VIZ)
    def _categorical_vs_continuous_violin_plot(x, y, data):
        # when x is categorical, we set the order based on counts of each category
        vc = data[x].value_counts()
        if len(vc.keys()) > 10:
            _log_plot_high_cardinality_warning(x, len(vc.keys()))
            idxes = vc[:10].index
        else:
            idxes = vc.index
        seaborn.violinplot(x=x, y=y, data=data[data[x].isin(idxes)], order=idxes)

    @staticmethod
    @runtime_dependency(module="seaborn", install_from=OptionalDependency.VIZ)
    def _ordinal_vs_continuous_violin_plot(x, y, data):
        # when x is ordinal, we want to get a natural order to x values
        vals = np.array(data[x].values)
        sorted_x = list(np.sort(vals))
        # get the frequency of each distinct element in the list using a dictionary.
        freq = {}
        for items in sorted_x:
            freq[items] = sorted_x.count(items)
        if len(freq) > MAX_DISPLAY_VALUES:
            _log_plot_high_cardinality_warning(x, len(freq))
            idxes = list(freq.keys())[:10]
        else:
            idxes = freq.keys()
        seaborn.violinplot(x=x, y=y, data=data[data[x].isin(idxes)], order=idxes)

    @staticmethod
    @runtime_dependency(module="seaborn", install_from=OptionalDependency.VIZ)
    def _categorical_vs_continuous_horizontal_violin_plot(x, y, data):
        vc = data[x].value_counts()
        if len(vc.keys()) > 10:
            _log_plot_high_cardinality_warning(x, len(vc.keys()))
            idxes = vc[:10].index
        else:
            idxes = vc.index
        seaborn.violinplot(
            x=y, y=x, data=data[data[x].isin(idxes)], order=idxes, orient="h"
        )

    @staticmethod
    @runtime_dependency(module="seaborn", install_from=OptionalDependency.VIZ)
    def _ordinal_vs_continuous_horizontal_violin_plot(x, y, data):
        # when x is ordinal, we want to get a natural order to x values
        vals = np.array(data[x].values)
        sorted_x = list(np.sort(vals))
        # get the frequency of each distinct element in the list using a dictionary.
        freq = {}
        for items in sorted_x:
            freq[items] = sorted_x.count(items)
        if len(freq) > MAX_DISPLAY_VALUES:
            _log_plot_high_cardinality_warning(x, len(freq))
            idxes = list(freq.keys())[:10]
        else:
            idxes = freq.keys()
        seaborn.violinplot(
            x=y, y=x, data=data[data[x].isin(idxes)], order=idxes, orient="h"
        )

    @staticmethod
    @runtime_dependency(module="seaborn", install_from=OptionalDependency.VIZ)
    def _categorical_vs_continuous_box_plot(x, y, data):
        vc = data[x].value_counts()
        if len(vc.keys()) > 10:
            _log_plot_high_cardinality_warning(x, len(vc.keys()))
            idxes = vc[:10].index
        else:
            idxes = vc.index
        seaborn.boxplot(x=x, y=y, data=data[data[x].isin(idxes)], order=idxes)

    @staticmethod
    @runtime_dependency(module="seaborn", install_from=OptionalDependency.VIZ)
    def _ordinal_vs_continuous_box_plot(x, y, data):
        # when x is ordinal, we want to get a natural order to x values
        vals = np.array(data[x].values)
        sorted_x = list(np.sort(vals))
        # get the frequency of each distinct element in the list using a dictionary.
        freq = {}
        for items in sorted_x:
            freq[items] = sorted_x.count(items)
        if len(freq) > MAX_DISPLAY_VALUES:
            _log_plot_high_cardinality_warning(x, len(freq))
            idxes = list(freq.keys())[:10]
        else:
            idxes = freq.keys()
        seaborn.boxplot(x=x, y=y, data=data[data[x].isin(idxes)], order=idxes)

    @staticmethod
    @runtime_dependency(module="seaborn", install_from=OptionalDependency.VIZ)
    def _ordinal_vs_continuous_horizontal_box_plot(x, y, data):
        # when x is ordinal, we want to get a natural order to x values
        vals = np.array(data[x].values)
        sorted_x = list(np.sort(vals))
        # get the frequency of each distinct element in the list using a dictionary.
        freq = {}
        for items in sorted_x:
            freq[items] = sorted_x.count(items)
        if len(freq) > MAX_DISPLAY_VALUES:
            _log_plot_high_cardinality_warning(x, len(freq))
            idxes = list(freq.keys())[:10]
        else:
            idxes = freq.keys()
        seaborn.boxplot(
            x=y, y=x, data=data[data[x].isin(idxes)], order=idxes, orient="h"
        )

    @staticmethod
    @runtime_dependency(module="seaborn", install_from=OptionalDependency.VIZ)
    def _count_plot(x, hue, data, yscale=None):
        if not yscale:
            _log_yscale_not_set()
        # get the copy of data and convert categorical data type to object
        data = data.copy()
        data[hue] = data[hue].astype("object")
        # get cardinality of categorical values
        cat_cardi = data[hue].value_counts()
        if len(cat_cardi) > 5:
            top_categoricals = cat_cardi[:5].index
            # modify the data to replace non top 5 categorical values to be the same value
            data[hue] = np.where(
                data[hue].isin(top_categoricals), data[hue], "all_other_categories"
            )
            cat_index = data[hue].value_counts().index
        else:
            cat_index = cat_cardi.index
        # get cardinality of ordinal values
        ordi_cardi = data[x].value_counts()
        if len(ordi_cardi) > 10:
            # bin the values and sort from small to large
            data[x] = pd.cut(data[x], 10, precision=0)
            data[x] = data[x].apply(lambda k: pd.Interval(int(k.left), int(k.right)))
        g = seaborn.countplot(x=x, hue=hue, data=data, hue_order=cat_index)
        if yscale:
            g.set_yscale(yscale)

    @staticmethod
    @runtime_dependency(module="seaborn", install_from=OptionalDependency.VIZ)
    def _ordinal_count_plot(x, data, yscale=None):
        if not yscale:
            _log_yscale_not_set()
        data = data.copy()
        ordi_cardi = data[x].value_counts()
        if len(ordi_cardi) > 20:
            intervals = pd.cut(data[x], 20, precision=0)
            intervals = intervals.apply(
                lambda k: pd.Interval(int(k.left), int(k.right))
            )
        else:
            intervals = data[x]
        g = seaborn.countplot(x=intervals, color="#1f77b4")
        if yscale:
            g.set_yscale(yscale)

    @staticmethod
    @runtime_dependency(module="seaborn", install_from=OptionalDependency.VIZ)
    def _ordinal_vs_constant_count_plot(x, hue, data, yscale=None):
        if not yscale:
            _log_yscale_not_set()
        data = data.copy()
        intervals = pd.cut(data[x], 20, precision=0)
        intervals = intervals.apply(lambda k: pd.Interval(int(k.left), int(k.right)))
        g = seaborn.countplot(x=intervals, hue=hue, data=data)
        if yscale:
            g.set_yscale(yscale)

    @staticmethod
    @runtime_dependency(module="seaborn", install_from=OptionalDependency.VIZ)
    def _single_column_count_plot(x, data, yscale=None):
        if not yscale:
            _log_yscale_not_set()
        order = data[x].value_counts().iloc[:24].index
        g = seaborn.countplot(x=x, data=data, order=order)
        if yscale:
            g.set_yscale(yscale)

    @staticmethod
    @runtime_dependency(module="IPython", install_from=OptionalDependency.NOTEBOOK)
    @runtime_dependency(module="folium", install_from=OptionalDependency.VIZ)
    def _folium_map(x, data):
        import folium.plugins
        df = LatLonDetector.extract_x_y(data[x])
        lat_min, lat_max, long_min, long_max = (
            min(df.Y),
            max(df.Y),
            min(df.X),
            max(df.X),
        )
        m = folium.Map(tiles="Stamen Terrain", zoom_control=False)

        folium.plugins.HeatMap(df[["Y", "X"]]).add_to(m)
        m.fit_bounds([[lat_min, long_min], [lat_max, long_max]])

        from IPython.core.display import display

        display(m)

    @staticmethod
    @runtime_dependency(module="seaborn", install_from=OptionalDependency.VIZ)
    def _single_pdf(x, y, data):
        seaborn.kdeplot(data[x], shade=True, shade_lowest=False)

    @staticmethod
    @runtime_dependency(module="seaborn", install_from=OptionalDependency.VIZ)
    def _multiple_pdf(x, y, data):
        colors = dict(mcolors.BASE_COLORS, **mcolors.CSS4_COLORS)
        hues = [
            colors[x]
            for x in colors.keys()
            if isinstance(colors[x], str) and colors[x].startswith("#")
        ]

        for i, cat in enumerate(list(data[y].unique())):
            s = data.loc[data[y] == cat][x]
            color = random.choice(hues)
            seaborn.kdeplot(s, color=color, shade=True, shade_lowest=False, label=cat)
        plt.xlabel(x)
        plt.ylabel(y)

    @runtime_dependency(module="seaborn", install_from=OptionalDependency.VIZ)
    def _matplot(self, plot_method, figsize=(4, 3), **kwargs):

        plt.style.use("seaborn-white")

        plt.rc("xtick", labelsize="x-small")
        plt.rc("ytick", labelsize="x-small")
        plt.rc("font", size=8)
        plt.rc("figure", dpi=144)

        fig = plt.figure(figsize=figsize)

        #
        # generate a title for the plot
        #
        text = '{}, "{}" ({})'.format(
            plot_method.__name__.upper(), self.x, self.feature_types[self.x].type
        )
        if self.y:
            text = '{} vs "{}" ({})'.format(
                text, self.y, self.feature_types[self.y].type
            )

        plt.title(text, y=1.08)
        plt.grid(linestyle="dotted")

        # draw a 45 degree dotted or dashed line indicating equality when plot method is scatter plots and the span of x equals the span of y
        if (
            plot_method is plt.scatter or plot_method is Plotting._gaussian_heatmap
        ) and (self.df[kwargs["x"]].values.ptp() == self.df[kwargs["y"]].values.ptp()):
            Plotting._add_identity(fig.axes[0], color="grey", ls="--")

        plot_method(**kwargs, data=self.df)

        # set x and y axis label
        if self.y:
            plt.ylabel(self.y)
        if self.x:
            plt.xlabel(self.x)

        # rename the y-axis label and x-axis label when "count" is the y-axis label
        if self.y == "count":
            plt.xlabel("Column: {} values ".format(self.x))
            plt.ylabel("instance count")

        # add y-axis label as "count" when plot type is hist
        if plot_method is plt.hist:
            plt.ylabel("instance count")
            # add tickmark on x-axis to see labeled values on x-axis in historgram. It has 30 intervals because that's the most tickmarks on graph can fit.
            plt.xticks(
                np.arange(
                    min(self.df[kwargs["x"]].values),
                    max(self.df[kwargs["x"]].values) + 1,
                    (
                        max(self.df[kwargs["x"]].values)
                        - min(self.df[kwargs["x"]].values)
                    )
                    / 30,
                )
            )
        # override y-axis label as "count" when plot type is _count_plot or countplot
        if plot_method is Plotting._count_plot or plot_method is seaborn.countplot:
            plt.ylabel("count")

        plt.xticks(rotation=90)

    def _generic_plot(self, plot_method, **kwargs):
        plot_method(**kwargs, data=self.df)

    @runtime_dependency(module="seaborn", install_from=OptionalDependency.VIZ)
    def _get_plot_method(self):

        #
        # combos contains a dictionary with the key being a composite of the x and y types, the value will
        # always be a list, possibly and empty list, indicating no match for combination
        #
        #

        combos = defaultdict(list)

        combos[
            Plotting._build_plot_key(CategoricalTypedFeature, ContinuousTypedFeature)
        ] = [
            (
                self._matplot,
                Plotting._categorical_vs_continuous_violin_plot,
                {"x": self.x, "y": self.y},
            ),
            (
                self._matplot,
                Plotting._categorical_vs_continuous_box_plot,
                {"x": self.x, "y": self.y},
            ),
        ]
        combos[
            Plotting._build_plot_key(OrdinalTypedFeature, ContinuousTypedFeature)
        ] = [
            (
                self._matplot,
                Plotting._ordinal_vs_continuous_violin_plot,
                {"x": self.x, "y": self.y},
            ),
            (
                self._matplot,
                Plotting._ordinal_vs_continuous_box_plot,
                {"x": self.x, "y": self.y},
            ),
        ]
        combos[
            Plotting._build_plot_key(ContinuousTypedFeature, OrdinalTypedFeature)
        ] = [
            (
                self._matplot,
                Plotting._ordinal_vs_continuous_horizontal_violin_plot,
                {"x": self.y, "y": self.x},
            ),
            (
                self._matplot,
                Plotting._ordinal_vs_continuous_horizontal_box_plot,
                {"x": self.y, "y": self.x},
            ),
        ]
        combos[
            Plotting._build_plot_key(ContinuousTypedFeature, CategoricalTypedFeature)
        ] = [(self._matplot, Plotting._multiple_pdf, {"x": self.x, "y": self.y})]
        combos[
            Plotting._build_plot_key(ConstantTypedFeature, ContinuousTypedFeature)
        ] = [(self._matplot, seaborn.barplot, {"x": self.x, "y": self.y})]
        combos[
            Plotting._build_plot_key(ContinuousTypedFeature, ConstantTypedFeature)
        ] = [(self._matplot, Plotting._single_pdf, {"x": self.x, "y": self.y})]
        combos[Plotting._build_plot_key(ConstantTypedFeature, DiscreteTypedFeature)] = [
            (self._matplot, seaborn.barplot, {"x": self.x, "y": self.y})
        ]
        combos[Plotting._build_plot_key(DiscreteTypedFeature, ConstantTypedFeature)] = [
            (
                self._matplot,
                Plotting._ordinal_vs_constant_count_plot,
                {"x": self.x, "hue": self.y, "yscale": self.yscale},
            )
        ]
        combos[
            Plotting._build_plot_key(DateTimeTypedFeature, ContinuousTypedFeature)
        ] = [
            (
                self._matplot,
                plt.scatter,
                {
                    "x": self.x,
                    "y": self.y,
                    "s": pi / 10 * (matplotlib.rcParams["lines.markersize"] ** 2),
                    "edgecolor": "white",
                    "linewidths": "0.1",
                },
            )
        ]
        combos[Plotting._build_plot_key(DateTimeTypedFeature, OrdinalTypedFeature)] = [
            (
                self._matplot,
                plt.scatter,
                {"x": self.x, "y": self.y, "edgecolor": "white", "linewidths": "0.1"},
            )
        ]
        combos[
            Plotting._build_plot_key(ContinuousTypedFeature, ContinuousTypedFeature)
        ] = [
            (
                self._matplot,
                Plotting._gaussian_heatmap,
                {
                    "x": self.x,
                    "y": self.y,
                    "s": pi / 10 * (matplotlib.rcParams["lines.markersize"] ** 2),
                },
            ),
            (
                self._matplot,
                plt.scatter,
                {"x": self.x, "y": self.y, "edgecolor": "white", "linewidths": "0.1"},
            ),
        ]
        combos[Plotting._build_plot_key(OrdinalTypedFeature, OrdinalTypedFeature)] = [
            (
                self._matplot,
                seaborn.scatterplot,
                {
                    "x": self.x,
                    "y": self.y,
                    "s": pi / 10 * (matplotlib.rcParams["lines.markersize"] ** 2),
                    "edgecolor": "white",
                    "linewidths": "0.1",
                },
            )
        ]
        combos[Plotting._build_plot_key(OrdinalTypedFeature, DiscreteTypedFeature)] = [
            (self._matplot, seaborn.countplot, {"x": self.x, "hue": self.y})
        ]
        combos[
            Plotting._build_plot_key(OrdinalTypedFeature, CategoricalTypedFeature)
        ] = [
            (
                self._matplot,
                Plotting._count_plot,
                {"x": self.x, "hue": self.y, "yscale": self.yscale},
            )
        ]
        combos[
            Plotting._build_plot_key(CategoricalTypedFeature, OrdinalTypedFeature)
        ] = [
            (
                self._matplot,
                Plotting._count_plot,
                {"x": self.y, "hue": self.x, "yscale": self.yscale},
            )
        ]
        combos[Plotting._build_plot_key(DiscreteTypedFeature, OrdinalTypedFeature)] = [
            (
                self._matplot,
                seaborn.countplot,
                {
                    "x": self.x,
                    "hue": self.y,
                    "order": self.df[self.x]
                    .value_counts(ascending=True)
                    .iloc[:10]
                    .index,
                },
            )
        ]
        combos[
            Plotting._build_plot_key(DiscreteTypedFeature, CategoricalTypedFeature)
        ] = [
            (
                self._matplot,
                seaborn.countplot,
                {
                    "x": self.x,
                    "hue": self.y,
                    "order": self.df[self.x].value_counts().iloc[:10].index,
                },
            )
        ]
        combos[Plotting._build_plot_key(DateTimeTypedFeature, DateTimeTypedFeature)] = [
            (
                self._matplot,
                plt.scatter,
                {
                    "x": self.x,
                    "y": self.y,
                    "s": pi / 10 * (matplotlib.rcParams["lines.markersize"] ** 2),
                    "edgecolor": "white",
                    "linewidths": "0.1",
                },
            )
        ]

        combos[Plotting._build_plot_key(ContinuousTypedFeature, None)] = [
            (self._matplot, plt.hist, {"x": self.x})
        ]
        combos[Plotting._build_plot_key(CategoricalTypedFeature, None)] = [
            (
                self._matplot,
                Plotting._single_column_count_plot,
                {"x": self.x, "yscale": self.yscale},
            )
        ]
        combos[Plotting._build_plot_key(OrdinalTypedFeature, None)] = [
            (
                self._matplot,
                Plotting._ordinal_count_plot,
                {"x": self.x, "yscale": self.yscale},
            )
        ]
        combos[Plotting._build_plot_key(ConstantTypedFeature, None)] = [
            (self._matplot, seaborn.countplot, {"x": self.x})
        ]
        combos[Plotting._build_plot_key(DateTimeTypedFeature, None)] = [
            (self._matplot, plt.hist, {"x": self.x, "bins": 10, "color": "#1f77b4"})
        ]
        combos[Plotting._build_plot_key(GISTypedFeature, None)] = [
            (self._generic_plot, Plotting._folium_map, {"x": self.x})
        ]

        y_type_name = None if self.y_type is None else self.y_type.__class__
        keys_to_check = list(
            [Plotting._build_plot_key(self.x_type.__class__, y_type_name)]
        )

        new_x_type = Plotting._change_type(self.x_type)
        new_y_type = Plotting._change_type(self.y_type)
        keys_to_check.append(Plotting._build_plot_key(new_x_type, y_type_name))
        keys_to_check.append(
            Plotting._build_plot_key(self.x_type.__class__, new_y_type)
        )
        keys_to_check.append(Plotting._build_plot_key(new_x_type, new_y_type))
        for key in keys_to_check:
            if key in combos and combos[key]:
                assert isinstance(combos[key][0], tuple)
                return combos[key]

        if y_type_name is not None:
            raise NotImplementedError(
                "Plotting for the feature combination ({0} vs {1}) is not yet supported.".format(
                    self.x_type.meta_data["type"], self.y_type.meta_data["type"]
                )
            )
        else:
            raise NotImplementedError(
                "Plotting for feature type {0} is not supported".format(
                    self.x_type.meta_data["type"]
                )
            )

    def _change_type(feature_type):
        if feature_type is None:
            return None

        return (
            DiscreteTypedFeature
            if isinstance(feature_type, DiscreteTypedFeature)
            or isinstance(feature_type, CreditCardTypedFeature)
            or isinstance(feature_type, ZipcodeTypedFeature)
            or isinstance(feature_type, OrdinalTypedFeature)
            else feature_type.__class__
        )



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from __future__ import print_function, absolute_import, division

import base64
from io import BytesIO
import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.lines as mlines
from matplotlib.ticker import FormatStrFormatter
import numpy as np
import math
from ads.common import logger
from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)
import itertools
import pandas as pd

MAX_TITLE_LEN = 20
MAX_LEGEND_LEN = 10
MAX_PLOTS_PER_ROW = 2
# Maximum class number evaluation plotting supporting for multiclass problems
MAX_PLOTTING_CLASSES = 10
# Maximum characters in class label able to be shown without being truncated
MAX_CHARACTERS_LEN = 13


def _fig_to_html(fig):
    tmpfile = BytesIO()
    fig.savefig(tmpfile, format="png")
    encoded = base64.b64encode(tmpfile.getvalue()).decode("utf-8")

    html = "<img src='data:image/png;base64,{}'>".format(encoded)
    return html


class EvaluationPlot:
    """EvaluationPlot holds data and methods for plots and it used to output them

    Attributes
    ----------
    baseline (bool):
        whether to plot the null model or zero information model
    baseline_kwargs (dict):
        keyword arguments for the baseline plot
    color_wheel (dict):
        color information used by the plot
    font_sz (dict):
        dictionary of plot methods
    perfect (bool):
        determines whether a "perfect" classifier curve is displayed
    perfect_kwargs (dict):
        parameters for the perfect classifier for precision/recall curves
    prob_type (str):
        model type, i.e. classification or regression

    Methods
    -------
    get_legend_labels(legend_labels)
        Renders the legend labels on the plot
    plot(evaluation, plots, num_classes, perfect, baseline, legend_labels)
        Generates the evalation plot
    """

    # dict of plot methods
    font_sz = {
        "xl": 16,  # Plot type title
        "l": 14,  # Individual plot title (name of model)
        "m": 12,  # Axis titles
        "s": 10,  # Axis labels
        "xs": 8,
    }  # test within the plot

    baseline_kwargs = {"ls": "--", "c": ".2"}  # lw = ??
    perfect_kwargs = {"label": "Perfect Classifier", "ls": "--", "color": "gold"}
    perfect = None
    baseline = None
    prob_type = None
    color_wheel = ["teal", "blueviolet", "forestgreen", "peru", "y", "dodgerblue", "r"]

    _pretty_titles_map = {
        "normalized_confusion_matrix": "Normalized Confusion Matrix",
        "lift_chart": "Lift Chart",
        "gain_chart": "Gain Chart",
        "ks_statistics": "KS Statistics",
        "residuals_qq": "Residuals Q-Q Plot",
        "residuals_vs_predicted": "Residuals vs Predicted",
        "residuals_vs_observed": "Residuals vs Observed",
        "observed_vs_predicted": "Observed vs Predicted",
        "precision_by_label": "Precision by Label",
        "recall_by_label": "Recall by Label",
        "f1_by_label": "F1 by Label",
        "jaccard_by_label": "Jaccard by Label",
        "pr_curve": "PR Curve",
        "roc_curve": "ROC Curve",
        "pr_and_roc_curve": "PR Curve, ROC Curve",
        "lift_and_gain_chart": "Lift Chart, Gain Chart",
    }

    _ugly_titles_map = {v: k for k, v in _pretty_titles_map.items()}

    double_overlay_plots = ["pr_and_roc_curve", "lift_and_gain_chart"]
    single_overlay_plots = ["lift_chart", "gain_chart", "roc_curve", "pr_curve"]

    _bin_plots = [
        "pr_curve",
        "roc_curve",
        "lift_chart",
        "gain_chart",
        "normalized_confusion_matrix",
    ]
    _multi_plots = [
        "normalized_confusion_matrix",
        "roc_curve",
        "pr_curve",
        "precision_by_label",
        "recall_by_label",
        "f1_by_label",
        "jaccard_by_label",
    ]
    _reg_plots = [
        "observed_vs_predicted",
        "residuals_qq",
        "residuals_vs_predicted",
        "residuals_vs_observed",
    ]

    # list of detailed descriptions of each plot type for every classification type, can be extended when adding more metrics
    _bin_plots_details = """
    In pattern recognition, information retrieval and binary classification, precision (also called positive predictive value)
    is the fraction of relevant instances among the retrieved instances, while recall (also known as sensitivity) is the
    fraction of relevant instances that have been retrieved over the total amount of relevant instances. \n
    A receiver operating characteristic curve, or ROC curve, is a graphical plot that illustrates the diagnostic ability of a
    binary classifier system as its discrimination threshold is varied. \n
    In data mining and association rule learning, lift is a measure of the performance of a targeting model (association rule)
    at predicting or classifying cases as having an enhanced response (with respect to the population as a whole), measured
    against a random choice targeting model. \n
    A gain graph is a graph whose edges are labelled 'invertibly', or 'orientably', by elements of a group G. \n
    In the field of machine learning and specifically the problem of statistical classification, a confusion matrix, also known
    as an error matrix, is a specific table layout that allows visualization of the performance of an algorithm, typically a
    supervised learning one (in unsupervised learning it is usually called a matching matrix).
    """
    # Removed for now:
    # Kuiper 's test (ks_statistics) is used in statistics to test that whether a given distribution, or family of
    # distributions, is contradicted by evidence from a sample of data. \n

    _multi_plots_details = """
    In the field of machine learning and specifically the problem of statistical classification, a confusion matrix, also known
    as an error matrix, is a specific table layout that allows visualization of the performance of an algorithm, typically a
    supervised learning one (in unsupervised learning it is usually called a matching matrix). \n
    A receiver operating characteristic curve, or ROC curve, is a graphical plot that illustrates the diagnostic ability of a
    binary classifier system as its discrimination threshold is varied. \n
    In pattern recognition, information retrieval and binary classification, precision (also called positive predictive value)
    is the fraction of relevant instances among the retrieved instances, while recall (also known as sensitivity) is the
    fraction of relevant instances that have been retrieved over the total amount of relevant instances. \n
    In statistical analysis of binary classification, the F1 score (also F-score or F-measure) is a measure of a test's accuracy.
    It considers both the precision p and the recall r of the test to compute the score: p is the number of correct positive results
    divided by the number of all positive results returned by the classifier, and r is the number of correct positive results divided
    by the number of all relevant samples (all samples that should have been identified as positive). The F1 score is the harmonic mean
    of the precision and recall, where an F1 score reaches its best value at 1 (perfect precision and recall) and worst at 0. \n
    The Jaccard index, also known as Intersection over Union and the Jaccard similarity coefficient, is a statistic used for gauging
    the similarity and diversity of sample sets. The Jaccard coefficient measures similarity between finite sample sets, and is defined
    as the size of the intersection divided by the size of the union of the sample sets
    """
    _reg_plots_details = """
    In statistics, a Q–Q (quantile-quantile) plot is a probability plot, which is a graphical method for comparing two probability distributions
    by plotting their quantiles against each other. \n
    In statistics and optimization, errors and residuals are two closely related and easily confused measures of the deviation of an observed value
    of an element of a statistical sample from its "theoretical value". The error (or disturbance) of an observed value is the deviation of the
    observed value from the (unobservable) true value of a quantity of interest (for example, a population mean), and the residual of an observed
    value is the difference between the observed value and the estimated value of the quantity of interest (for example, a sample mean).
    """

    @classmethod
    def _get_formatted_title(cls, title, max_len=MAX_TITLE_LEN):
        return title if len(title) < max_len + 3 else title[:max_len] + "..."

    @classmethod
    def get_legend_labels(cls, legend_labels):
        """Gets the legend labels, resolves any conflicts such as length, and renders
        the labels for the plot

        Parameters
        ----------
        legend_labels (dict):
            key/value dictionary containing legend label data

        Returns
        -------
        Nothing

        Examples
        --------

        EvaluationPlot.get_legend_labels({'class_0': 'green', 'class_1': 'yellow', 'class_2': 'red'})
        """

        @runtime_dependency(module="IPython", install_from=OptionalDependency.NOTEBOOK)
        @runtime_dependency(
            module="ipywidgets", object="HTML", install_from=OptionalDependency.NOTEBOOK
        )
        def render_legend_labels(label_dict):
            encodings = pd.DataFrame(
                pd.Series(label_dict, index=label_dict.keys()),
                columns=["Shortened labels"],
            )
            from IPython.core.display import display, HTML

            display(
                HTML(
                    encodings.style.format(precision=4)
                    .set_properties(**{"text-align": "center"})
                    .set_table_styles(
                        [dict(selector="", props=[("text-align", "center")])]
                    )
                    .set_table_attributes("class=table")
                    .set_caption(
                        '<div align="left"><b style="font-size:20px;">'
                        + "Legend for labels of the target feature:</b></div>"
                    )
                    .to_html()
                )
            )

        if legend_labels is not None:
            # CAUTION: cls.classes is a list of strings. Make sure users know that labels are all converted to strings.
            if isinstance(legend_labels, dict) and set(cls.classes).issubset(
                set(legend_labels.keys())
            ):
                render_legend_labels(legend_labels)
                cls.legend_labels = legend_labels
            else:
                logger.error(
                    "The provided `legend_labels` is either not a Python dict or does not possess all possible class labels."
                )
            return

        # try to remove leading words
        def _check_for_redundant_words(label_vec, prefix, max_len=MAX_LEGEND_LEN):
            words_found = False
            classes = [lab for lab in label_vec]
            while len(set([lab.split()[0] for lab in classes])) <= 1:
                # remove that word from vec, and add it to prefix
                additional_prefix = classes[0].split()[0]
                prefix = (
                    prefix[:-3] + additional_prefix + " ..."
                    if words_found
                    else additional_prefix + " ..."
                )
                classes = [lab[len(additional_prefix) + 1 :] for lab in classes]
                words_found = True

            classes = (
                classes if words_found else [label[max_len:] for label in label_vec]
            )
            return classes, prefix

        # returns mapping from real labels to psuedo-labels, when psuedo-labels are not the first X letter
        def _resolve_conflict(label_vec, prefix, max_len=MAX_LEGEND_LEN):
            classes, prefix = _check_for_redundant_words(label_vec, prefix)
            label_dict = _get_labels(classes, max_len=max_len)
            resolved = {}
            for orig, new in label_dict.items():
                resolved[prefix[:-3] + orig] = "..." + new if new[:3] != "..." else new
            return resolved

        # returns mapping from provided list of strings, to short, unique substrings
        def _get_labels(classes, max_len=MAX_LEGEND_LEN):
            conflict_dict = {}
            for label in classes:
                prefix = label if len(label) < max_len + 3 else label[:max_len] + "..."
                if conflict_dict.get(prefix, None) is None:
                    conflict_dict[prefix] = [label]
                else:
                    conflict_dict[prefix].append(label)
            out = {}
            for k, v in conflict_dict.items():
                if len(v) == 1:
                    out[v[0]] = k
                else:
                    resolved = _resolve_conflict(v, k, max_len=MAX_LEGEND_LEN)
                    out.update(resolved)
            return out

        cls.legend_labels = _get_labels(cls.classes)
        if set(cls.legend_labels.keys()) != set(cls.legend_labels.values()):
            logger.info(
                f"Class labels greater than {MAX_CHARACTERS_LEN} characters have been truncated. "
                "Use the `legend_labels` parameter to define labels."
            )
            render_legend_labels(cls.legend_labels)

    # evaluation is a DataFrame with models as columns and metrics as rows
    @classmethod
    def plot(
        cls,
        evaluation,
        plots,
        num_classes,
        perfect=False,
        baseline=True,
        legend_labels=None,
    ):
        """Generates the evaluation plot

        Parameters
        ----------
        evaluation (DataFrame):
            DataFrame with models as columns and metrics as rows.
        plots (str):
            The plot type based on class attribute `prob_type`.
        num_classes (int):
            The number of classes for the model.
        perfect (bool, optional):
            Whether to display the curve of a perfect classifier. Default value is `False`.
        baseline (bool, optional):
            Whether to display the curve of the baseline, featureless model. Default value is `True`.
        legend_labels (dict, optional):
            Legend labels dictionary. Default value is `None`. If legend_labels not specified class names will be used for plots.

        Returns
        -------
        Nothing
        """

        cls.perfect = perfect
        cls.baseline = baseline
        # get plots to show
        if num_classes == 2:
            cls.prob_type = "_bin"
        elif num_classes > 2:
            cls.prob_type = "_multi"
        else:
            cls.prob_type = "_reg"
        plot_details = getattr(cls, cls.prob_type + "_plots_details")
        if plots is None:
            plots = getattr(cls, cls.prob_type + "_plots")
            logger.info(
                "Showing plot types: {}.".format(
                    ", ".join(
                        [
                            "{}".format(EvaluationPlot._pretty_titles_map[str(p)])
                            for p in plots
                        ]
                    ),
                    ", ".join(["{}".format(x) for x in map(str, plots)]),
                )
            )
            logger.info(plot_details)

        if cls.prob_type == "_bin":
            if "lift_chart" in plots and "gain_chart" in plots:
                plots.remove("lift_chart")
                plots.remove("gain_chart")
                plots.insert(0, "lift_and_gain_chart")

            if "roc_curve" in plots and "pr_curve" in plots:
                plots.remove("roc_curve")
                plots.remove("pr_curve")
                plots.insert(0, "pr_and_roc_curve")
        elif cls.prob_type == "_multi":
            if (
                "normalized_confusion_matrix" in plots
                and len(evaluation[evaluation.columns[0]]["classes"])
                >= MAX_PLOTTING_CLASSES
            ):
                logger.error(
                    f"Evaluation plotting is not yet supported for multiclass problems with {MAX_PLOTTING_CLASSES} or more classes."
                )
                plots = []
        classes = evaluation[evaluation.columns[0]]["classes"]
        if classes is not None:
            # CAUTION: class labels are converted to strings here.
            # If users are passing in legend_labels, they have to use strings as well.
            # Otherwise get_legend_labels() will complaint.
            # If users are not passing in legend_labels, get_legend_labels() generates them from cls.classes.
            # cls.legend_labels are assigned/created in cls.get_lengend_labels() and contain only strings as keys.
            cls.classes = [str(c) for c in classes]
            cls.get_legend_labels(legend_labels)

        mpl.style.use("default")
        html_raw = []
        for i, plot_type in enumerate(plots):
            fig_title, fig = None, None
            try:
                fig_title, ax_title = plt.subplots(1, 1, figsize=(18, 0.5), dpi=144)
                ax_title.text(
                    0.5,
                    0.5,
                    cls._pretty_titles_map[plot_type],
                    fontsize=16,
                    fontweight="semibold",
                    horizontalalignment="center",
                    verticalalignment="center",
                    transform=ax_title.transAxes,
                )
                ax_title.axis("off")
                html_raw.append(_fig_to_html(fig_title))
                if cls.prob_type == "_bin" and plot_type in cls.double_overlay_plots:
                    fig, ax = plt.subplots(1, 2, figsize=(12, 4.5), dpi=144)
                elif cls.prob_type == "_bin" and plot_type in ["roc_curve", "pr_curve"]:
                    fig, axs = plt.subplots(1, 2, figsize=(12, 4.5), dpi=144)
                    axs[1].axis("off")
                    ax = [axs[0]]
                elif cls.prob_type == "_bin" and plot_type in [
                    "lift_chart",
                    "gain_chart",
                ]:
                    fig, axs = plt.subplots(1, 2, figsize=(12, 4.5), dpi=144)
                    axs[1].axis("off")
                    ax = axs[0]
                else:
                    nrows = math.ceil(len(evaluation.columns) / MAX_PLOTS_PER_ROW)
                    fig, ax = plt.subplots(
                        nrows, MAX_PLOTS_PER_ROW, figsize=(10, 4 * nrows), dpi=144
                    )  # 10, 3.5
                    ax = ax.flatten()
                getattr(cls, "_" + plot_type)(ax, evaluation)
                fig.tight_layout()
                html_raw.append(_fig_to_html(fig))
            except KeyError as e:
                try:
                    if fig_title:
                        plt.close(fig=fig_title)
                    if fig:
                        plt.close(fig=fig)
                except:
                    pass
                logger.warning(
                    f"Evaluator was not able to plot "
                    f"{cls._pretty_titles_map.get(plot_type, plot_type)}, because the relevant "
                    f"metrics had complications. Ensure that `predict` and `predict_proba` "
                    f"are valid."
                )
        return html_raw

    @classmethod
    def _lift_and_gain_chart(cls, ax, evaluation):
        cls._lift_chart(ax[0], evaluation)
        cls._gain_chart(ax[1], evaluation)

    @classmethod
    def _lift_chart(cls, ax, evaluation):
        for mod_name, col in evaluation.iteritems():
            if col["y_score"] is not None:
                ax.plot(
                    col["percentages"][1:],
                    [1] + list(col["lift"]),
                    label=cls._get_formatted_title(mod_name),
                )
        if cls.baseline:
            ax.plot([-10, 110], [1, 1], **cls.baseline_kwargs)
        if cls.perfect:
            perf_idx = next(
                idx
                for idx, scores in enumerate(evaluation.loc["y_score"])
                if scores is not None
            )
            ax.plot(
                evaluation.loc["percentages"][perf_idx][1:],
                [1] + list(evaluation.loc["perfect_lift"][perf_idx]),
                **cls.perfect_kwargs,
            )
        ax.legend(loc="upper right", frameon=False)
        ax.set_xlabel("Percentage of Population", fontsize=12)
        ax.set_ylabel("Lift", fontsize=12)
        ax.set_title("Lift Chart", y=1.08, fontsize=14)
        ax.grid(linewidth=0.2, which="both")
        ax.set_xlim([-10, 110])

    @classmethod
    def _gain_chart(cls, ax, evaluation):
        for mod_name, col in evaluation.iteritems():
            if col["y_score"] is not None:
                ax.plot(
                    col["percentages"],
                    list(col["cumulative_gain"]),
                    label=cls._get_formatted_title(mod_name),
                )
        if cls.baseline:
            ax.plot([-10, 110], [-10, 110], **cls.baseline_kwargs)
        if cls.perfect:
            perf_idx = next(
                idx
                for idx, scores in enumerate(evaluation.loc["y_score"])
                if scores is not None
            )
            ax.plot(
                evaluation.loc["percentages"][perf_idx],
                evaluation.loc["perfect_gain"][perf_idx],
                **cls.perfect_kwargs,
            )
        ax.legend(loc="lower right", frameon=False)
        ax.set_xlabel("Percentage of Population", fontsize=12)
        ax.set_ylabel("Percentage of Positive Class", fontsize=12)
        ax.set_title("Gain Chart", y=1.08, fontsize=14)
        ax.grid(linewidth=0.2, which="both")
        ax.set_xlim([-10, 110])
        ax.set_ylim([-10, 110])

    @classmethod
    def _pr_and_roc_curve(cls, ax, evaluation):
        cls._pr_curve([ax[0]], evaluation)
        cls._roc_curve([ax[1]], evaluation)

    @classmethod
    def _pr_curve(cls, axs, evaluation):
        n_models = len(evaluation.columns)
        for i, ax in enumerate(axs):
            if i >= n_models:
                ax.axis("off")
                return
            if cls.prob_type == "_bin":
                for mod_name, col in evaluation.iteritems():
                    if col["y_score"] is not None:
                        ax.plot(
                            col["recall_values"],
                            col["precision_values"],
                            label="%s (Precision: %s)"
                            % (
                                cls._get_formatted_title(mod_name),
                                "{:.3f}".format(col["precision"]),
                            ),
                        )
                        ax.plot(
                            *col["pr_best_model_score"],
                            color=ax.get_lines()[-1].get_color(),
                            marker="*",
                        )
            else:
                model_name = evaluation.columns[i]
                mod = evaluation[model_name]
                if mod["y_score"] is not None:
                    for j, lab in enumerate(mod.classes):
                        # cls.legend_labels contains only strings as keys.
                        lab = str(lab)
                        ax.plot(
                            mod["recall_values"][j],
                            mod["precision_values"][j],
                            label="%s (Precision: %s)"
                            % (
                                cls.legend_labels[lab],
                                "{:.3f}".format(mod["precision_by_label"][j]),
                            ),
                        )
                        ax.plot(
                            *mod["pr_best_model_score"][j],
                            color=ax.get_lines()[-1].get_color(),
                            marker="*",
                        )

            ax.set_xlabel("Recall", fontsize=12)
            ax.set_ylabel("Precision", fontsize=12)
            ax.set_title("Precision Recall Curve", y=1.08, fontsize=14)
            ax.grid(linewidth=0.2, which="both")
            ax.set_xlim([-0.1, 1.1])
            ax.set_ylim([-0.1, 1.1])
            handles, labels = ax.get_legend_handles_labels()
            star = mlines.Line2D(
                [],
                [],
                color="black",
                marker="*",
                linestyle="None",
                markersize=5,
                label="Minimum Error Rate",
            )
            handles.append(star)
            labels.append("Minimum Error Rate")
            ax.legend(
                loc="upper right",
                labels=labels,
                handles=handles,
                frameon=False,
                fontsize="x-small",
            )

    @classmethod
    def _roc_curve(cls, axs, evaluation):
        n_models = len(evaluation.columns)
        for i, ax in enumerate(axs):
            if i >= n_models:
                ax.axis("off")
                return
            if cls.prob_type == "_bin":
                for mod_name, col in evaluation.iteritems():
                    if col["y_score"] is not None:
                        ax.plot(
                            col["false_positive_rate"],
                            col["true_positive_rate"],
                            label="%s (AUC: %s)"
                            % (
                                cls._get_formatted_title(mod_name),
                                "{:.3f}".format(col["auc"]),
                            ),
                        )
                        ax.plot(
                            *col["roc_best_model_score"],
                            color=ax.get_lines()[-1].get_color(),
                            marker="*",
                        )
            else:
                model_name = evaluation.columns[i]
                mod = evaluation[model_name]
                if mod["y_score"] is not None:
                    for j, lab in enumerate(mod.classes):
                        # cls.legend_labels contains only strings as keys.
                        lab = str(lab)
                        ax.plot(
                            mod["fpr_by_label"][j],
                            mod["tpr_by_label"][j],
                            label="%s (AUC: %s)"
                            % (cls.legend_labels[lab], "{:.3f}".format(mod["auc"][j])),
                        )
                        ax.plot(
                            *mod["roc_best_model_score"][j],
                            color=ax.get_lines()[-1].get_color(),
                            marker="*",
                        )
            if cls.baseline:
                ax.plot([-0.1, 1.1], [-0.1, 1.1], **cls.baseline_kwargs)
            ax.set_xlabel("False Positive Rate", fontsize=12)
            ax.set_ylabel("True Positive Rate", fontsize=12)
            ax.set_title("ROC Curve", y=1.08, fontsize=14)
            ax.grid(linewidth=0.2, which="both")
            ax.set_xlim([-0.1, 1.1])
            ax.set_ylim([-0.1, 1.1])
            handles, labels = ax.get_legend_handles_labels()
            star = mlines.Line2D(
                [],
                [],
                color="black",
                marker="*",
                linestyle="None",
                markersize=5,
                label="Youden's J Statistic",
            )
            handles.append(star)
            labels.append("Youden's J Statistic")
            ax.legend(
                loc="lower right",
                labels=labels,
                handles=handles,
                frameon=False,
                fontsize="x-small",
            )

    @classmethod
    def _ks_statistics(cls, axs, evaluation):
        n_models = len(evaluation.columns)
        for i, ax in enumerate(axs):
            if i >= n_models:
                ax.axis("off")
                return
            model_name = evaluation.columns[i]
            mod = evaluation[model_name]

            ax.set_title(model_name, fontsize=14)
            if mod["y_score"] is not None:
                ax.plot(
                    mod["ks_thresholds"],
                    mod["ks_pct1"],
                    lw=3,
                    label=mod["ks_labels"][0],
                )
                ax.plot(
                    mod["ks_thresholds"],
                    mod["ks_pct2"],
                    lw=3,
                    label=mod["ks_labels"][1],
                )
                if cls.baseline:
                    idx = np.where(mod["ks_thresholds"] == mod["max_distance_at"])[0][0]
                    ax.axvline(
                        mod["max_distance_at"],
                        *sorted([mod["ks_pct1"][idx], mod["ks_pct2"][idx]]),
                        label="KS Statistic: {:.3f} at {:.3f}".format(
                            mod["ks_statistic"], mod["max_distance_at"]
                        ),
                        linestyle="--",
                        color=".2",
                    )

            ax.set_xlim([0.0, 1.0])
            ax.set_ylim([0.0, 1.0])

            ax.set_xlabel("Threshold", fontsize=12)
            ax.set_ylabel("Percentage below threshold", fontsize=12)
            ax.tick_params(labelsize=10)
            ax.legend(loc="lower right", fontsize=8)

    @classmethod
    def _pretty_barh(
        cls, ax, x, y, axis_labels=None, title=None, axis_lim=None, plot_kwargs=None
    ):
        # cls.legend_labels contains only strings as keys.
        new_lab = [cls.legend_labels[str(item)] for item in x]
        ax.barh(
            new_lab,
            y,
            color=["teal", "blueviolet", "forestgreen", "peru", "y", "dodgerblue", "r"],
        )
        for j, v in enumerate(y):
            ax.annotate("{:.3f}".format(v), xy=(v / 2, j), va="center", ha="left")
        if axis_labels:
            if axis_labels[0]:
                ax.set_xlabel(axis_labels[0], fontsize=12)
            if axis_labels[1]:
                ax.set_ylabel(axis_labels[1], fontsize=12)
        if title:
            title = cls._get_formatted_title(title)
            ax.set_title(title, y=1.08, fontsize=14)
        if axis_lim:
            ax.set_xlim(axis_lim)

    @classmethod
    def _precision_by_label(cls, axs, evaluation):
        n_models = len(evaluation.columns)
        for i, ax in enumerate(axs):
            if i < n_models:
                col = evaluation.columns[i]
                cls._pretty_barh(
                    ax,
                    evaluation[col]["classes"],
                    evaluation[col]["precision_by_label"],
                    axis_lim=[0, 1],
                    axis_labels=["Precision", None],
                    title=col,
                )
            else:
                ax.axis("off")

    @classmethod
    def _recall_by_label(cls, axs, evaluation):
        n_models = len(evaluation.columns)
        for i, ax in enumerate(axs):
            if i < n_models:
                col = evaluation.columns[i]
                cls._pretty_barh(
                    ax,
                    evaluation[col]["classes"],
                    evaluation[col]["recall_by_label"],
                    axis_lim=[0, 1],
                    axis_labels=["Recall", None],
                    title=col,
                )
            else:
                ax.axis("off")

    @classmethod
    def _f1_by_label(cls, axs, evaluation):
        n_models = len(evaluation.columns)
        for i, ax in enumerate(axs):
            if i < n_models:
                col = evaluation.columns[i]
                cls._pretty_barh(
                    ax,
                    evaluation[col]["classes"],
                    evaluation[col]["f1_by_label"],
                    axis_lim=[0, 1],
                    axis_labels=["F1 Score", None],
                    title=col,
                )
            else:
                ax.axis("off")

    @classmethod
    def _jaccard_by_label(cls, axs, evaluation):
        n_models = len(evaluation.columns)
        for i, ax in enumerate(axs):
            if i < n_models:
                col = evaluation.columns[i]
                cls._pretty_barh(
                    ax,
                    evaluation[col]["classes"],
                    evaluation[col]["jaccard_by_label"],
                    axis_lim=[0, 1],
                    axis_labels=["Jaccard Score", None],
                    title=col,
                )
            else:
                ax.axis("off")

    @classmethod
    def _pretty_scatter(
        cls,
        ax,
        x,
        y,
        s=5,
        alpha=1.0,
        title=None,
        legend=False,
        axis_labels=None,
        axis_lim=None,
        grid=True,
        label=None,
        plot_kwargs=None,
    ):

        if plot_kwargs is None:
            plot_kwargs = {}
        ax.scatter(x, y, s=s, label=label, marker="o", alpha=alpha, **plot_kwargs)
        ax.xaxis.set_major_formatter(FormatStrFormatter("%.2f"))
        ax.yaxis.set_major_formatter(FormatStrFormatter("%.2f"))
        if legend:
            ax.legend(frameon=False)
        if axis_labels:
            ax.set_xlabel(axis_labels[0])
            ax.set_ylabel(axis_labels[1])
        if title:
            ax.set_title(title, y=1.08, fontsize=14)
        if grid:
            ax.grid(linewidth=0.2)
        if axis_lim:
            if axis_lim[0]:
                ax.set_xlim(axis_lim[0])
            if axis_lim[1]:
                ax.set_ylim(axis_lim[1])

    @classmethod
    def _top_2_features(cls, axs, evaluation):
        pass

    @classmethod
    def _residuals_qq(cls, axs, evaluation):
        n_models = len(evaluation.columns)
        for i, ax in enumerate(axs):
            if i >= n_models:
                ax.axis("off")
                return
            model_name = evaluation.columns[i]
            mod = evaluation[model_name]
            # getattr(ax, self.plot_method)(self.x, y, **self.plot_kwargs, color='#4a91c2', label=label)
            cls._pretty_scatter(
                ax,
                mod["norm_quantiles"],
                mod["residual_quantiles"],
                title=model_name,
                axis_lim=[(-2.7, 2.7), (-3.1, 3.1)],
                axis_labels=["Theoretical Quantiles", "Sample Quantiles"],
            )
            if cls.baseline:
                ax.plot((-100, 100), (-100, 100), **cls.baseline_kwargs)

    @classmethod
    def _residuals_vs_predicted(cls, axs, evaluation):
        n_models = len(evaluation.columns)
        for i, ax in enumerate(axs):
            if i >= n_models:
                ax.axis("off")
                return
            model_name = evaluation.columns[i]
            mod = evaluation[model_name]
            y_pred = np.asarray(mod["y_pred"])
            resid = np.asarray(mod["residuals"])
            cls._pretty_scatter(
                ax,
                y_pred,
                resid,
                s=4,
                alpha=0.5,
                title=model_name,
                axis_labels=["Predicted Values", "Residuals"],
            )
            x_lim = (
                y_pred.min() - y_pred.min() * 0.05,
                y_pred.max() + y_pred.max() * 0.05,
            )
            if cls.baseline:
                ax.plot(x_lim, (0, 0), **cls.baseline_kwargs)
            ax.set_xlim(x_lim)
            ax.xaxis.set_major_formatter(FormatStrFormatter("%.2f"))
            ax.yaxis.set_major_formatter(FormatStrFormatter("%.2f"))

    @classmethod
    def _residuals_vs_observed(cls, axs, evaluation):
        n_models = len(evaluation.columns)
        for i, ax in enumerate(axs):
            if i >= n_models:
                ax.axis("off")
                return
            model_name = evaluation.columns[i]
            mod = evaluation[model_name]
            y_true = np.asarray(mod["y_true"])
            cls._pretty_scatter(
                ax,
                y_true,
                mod["residuals"],
                s=4,
                alpha=0.5,
                title=model_name,
                axis_labels=["Observed Values", "Residuals"],
            )
            x_lim = (
                y_true.min() - y_true.min() * 0.05,
                y_true.max() + y_true.max() * 0.05,
            )
            if cls.baseline:
                ax.plot(x_lim, (0, 0), **cls.baseline_kwargs)
            ax.set_xlim(x_lim)
            ax.xaxis.set_major_formatter(FormatStrFormatter("%.2f"))
            ax.yaxis.set_major_formatter(FormatStrFormatter("%.2f"))

    @classmethod
    def _observed_vs_predicted(cls, axs, evaluation):
        n_models = len(evaluation.columns)
        for i, ax in enumerate(axs):
            if i >= n_models:
                ax.axis("off")
                return

            model_name = evaluation.columns[i]
            mod = evaluation[model_name]
            ax.scatter(mod["y_true"], mod["y_pred"], s=4, marker="o", alpha=0.5)

            y_true = np.asarray(mod["y_true"])

            yt_min = y_true.min()
            yt_max = y_true.max()

            x_lim = (yt_min - yt_min * 0.05, yt_max + yt_max * 0.05)
            if cls.baseline:
                ax.plot(x_lim, x_lim, **cls.baseline_kwargs)
            ax.set_xlabel("Observed Values")
            ax.set_ylabel("Predicted Values")
            ax.set_title(model_name, y=1.08, fontsize=14)
            ax.grid(linewidth=0.2)
            ax.set_xlim(x_lim)

            ax.xaxis.set_major_formatter(FormatStrFormatter("%.2f"))
            ax.yaxis.set_major_formatter(FormatStrFormatter("%.2f"))

    @classmethod
    def _normalized_confusion_matrix(cls, axs, evaluation):
        for model_num, ax in enumerate(axs):
            if model_num >= len(evaluation.columns):
                ax.axis("off")
                return
            model_name = evaluation.columns[model_num]
            mod = evaluation[model_name]
            if cls.prob_type == "_bin":
                labels = [str(lab == mod["positive_class"]) for lab in mod["classes"]]
            else:
                labels = cls.legend_labels.values()

            raw_cm = mod["raw_confusion_matrix"]
            cm = np.asarray(mod["confusion_matrix"])

            ax.set_title(
                "%s\n" % cls._get_formatted_title(model_name), y=1.08, fontsize=14
            )
            ax.imshow(cm, interpolation="nearest", cmap="BuGn")
            x_tick_marks = np.arange(len(labels))
            y_tick_marks = np.arange(len(labels))
            ax.set_xticks(x_tick_marks)
            ax.set_yticks(y_tick_marks)

            ax.set_xticklabels(labels, rotation=90, fontsize=10)
            ax.set_yticklabels(labels, fontsize=10)

            for i, j in itertools.product(
                range(raw_cm.shape[0]), range(raw_cm.shape[1])
            ):
                ax.text(
                    j,
                    i,
                    "%s [%s]" % (round(cm[i][j], 3), raw_cm[i, j]),
                    horizontalalignment="center",
                    verticalalignment="center",
                    rotation=45,
                    fontsize=max(3, 10 - max(raw_cm.shape[0], raw_cm.shape[1])),
                    color="white" if cm[i, j] > 0.5 else "black",
                )

            ax.set_ylabel("True label", fontsize=10)
            ax.set_xlabel("Predicted label", fontsize=10)
            ax.grid(False)



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

try:
    from mlx import LimeExplainer
    from mlx import FDExplainer, AleExplainer
    from mlx import PermutationImportance
    from mlx.whatif import WhatIf
except:
    pass
from ads.common import logger
from ads.dataset.helper import is_text_data

import pandas as pd
from ads.common.decorator.deprecate import deprecated


def _reset_index(x):
    assert isinstance(x, pd.DataFrame) or isinstance(x, pd.Series)
    return x.reset_index(drop=True)


def check_tabular_or_text(est, X):
    """
    Returns "text" if a text dataset, "tabular" otherwise.

    Parameters
    ----------
    est : ADSModel
        Model to explain.
    X : pandas.DataFrame
        Dataset.

    Return
    ------
    str
        "text" or "tabular"
    """
    return "text" if is_text_data(X) else "tabular"


@deprecated(
    details="Working with AutoML has moved from within ADS to working directly with the AutoMLx library. AutoMLx are preinstalled in conda pack automlx_p38_cpu_v2 and later, and can now be updated independently of ADS. AutoMLx documentation may be found at https://docs.oracle.com/en-us/iaas/tools/automlx/latest/html/multiversion/v23.1.1/index.html. Notebook examples are in Oracle's samples repository: https://github.com/oracle-samples/oci-data-science-ai-samples/tree/master/notebook_examples and a migration tutorial can be found at https://accelerated-data-science.readthedocs.io/en/latest/user_guide/model_training/automl/quick_start.html .",
    raise_error=True,
)
def init_lime_explainer(
    explainer,
    est,
    X_train,
    y_train,
    mode,
    class_names=None,
    selected_features=None,
    client=None,
    batch_size=16,
    surrogate_model="linear",
    num_samples=5000,
    exp_sorting="absolute",
    discretization="decile",
    scale_weight=True,
):
    """
    Initializes a local LIME Explainer. Also supports aggregate local
    explanations, which approximates a global behavior based on multiple local explanations.

    Supports both tabular and text datasets. The explainer is initialized to the defaults
    and can be updated with `MLXGlobalExplainer.configure_feature_importance()` or
    `MLXLocalExplainer.configure_local_explainer()`.

    Parameters
    ----------
    explainer : LimeExplainer, None
        If the explainer has previously been initialized, it can be passed in to avoid
        creating a new explainer object. If `None`, a new `LimeExplainer` instance will be created.
    est : ADSModel
        Model to explain.
    X_train : pandas.DataFrame
        Training dataset.
    y_train : pandas.DataFrame/Series
        Training labels.
    mode : str
        'classification' or 'regression'.
    class_names : list
        List of target names.
    selected_features : list[int], optional
        Pass the selected features information to the explainer. Defaults value is `None`.
    client : Dask Client
        Specifies that Dask Client object to use in MLX. If None, no parallelization.
    batch_size : int
        Number of local explanations that are batched and processed by each Dask worker
        in parallel.
    surrogate_model : str
        Surrogate model to approximate the local behavior of the ML model. Can be
        'linear' or 'decision_tree'.
    num_samples : int
        Number of samples the local explainer generates in the local neighborhood
        around the sample to explain to fit the surrogate model.
    exp_sorting : str
        Order of how to sort the feature importances. Can be 'absolute' or 'ordered'.
        Absolute ordering orders based on the absolute values, while ordered considers
        the sign of the feature importance values.
    discretizer : str
        Method to discretize continuous features in the local explainer. Supports 'decile',
        'quartile', 'entropy', and `None`. If `None`, the continuous feature values are
        used directly. If not None, each continuous feature is discretized and treated
        as a categorical feature.
    scale_weight : bool
        Normalizes the feature importance coefficients from the local explainer to sum to one.

    Return
    ------
    :class:`mlx.LimeExplainer`
    """
    if explainer is None:
        exp = LimeExplainer()
    else:
        if not isinstance(explainer, LimeExplainer):
            raise TypeError(
                "Invalid explainer provided to "
                "init_lime_explainer: {}".format(type(explainer))
            )
        exp = explainer
    exp_type = check_tabular_or_text(est, X_train)
    exp.set_config(
        type=exp_type,
        mode=mode,
        discretizer=discretization,
        client=client,
        batch_size=batch_size,
        scale_weight=scale_weight,
        surrogate_model=surrogate_model,
        num_samples=num_samples,
        exp_sorting=exp_sorting,
        kernel_width="dynamic",
    )
    exp.fit(
        est,
        X_train,
        y=y_train,
        target_names=class_names,
        selected_features=selected_features,
    )
    return exp


@deprecated(
    details="Working with AutoML has moved from within ADS to working directly with the AutoMLx library. AutoMLx are preinstalled in conda pack automlx_p38_cpu_v2 and later, and can now be updated independently of ADS. AutoMLx documentation may be found at https://docs.oracle.com/en-us/iaas/tools/automlx/latest/html/multiversion/v23.1.1/index.html. Notebook examples are in Oracle's samples repository: https://github.com/oracle-samples/oci-data-science-ai-samples/tree/master/notebook_examples and a migration tutorial can be found at https://accelerated-data-science.readthedocs.io/en/latest/user_guide/model_training/automl/quick_start.html .",
    raise_error=True,
)
def init_permutation_importance_explainer(
    explainer,
    est,
    X_train,
    y_train,
    mode,
    class_names=None,
    selected_features=None,
    client=None,
    random_state=42,
):
    """
    Initializes a Global Feature Permutation Importance Explainer.

    Supported for tabular datasets only.

    The explainer is initialized to the defaults and can be updated with
    MLXGlobalExplainer.configure_feature_importance().

    Parameters
    ----------
    explainer : PermutationImportance, None
        If the explainer has previously been initialized, it can be passed in to avoid
        creating a new explainer object. If `None`, a new `PermutationImportance` explainer
        will be created.
    est : ADSModel
        Model to explain.
    X_train : pandas.DataFrame
        Training dataset.
    y_train : pandas.DataFrame/Series
        Training labels.
    mode : str
        'classification' or 'regression'.
    class_names : list, optional
        List of target names. Default value is `None`
    selected_features : list[str], list[int], optional
        Pass the selected features list to the explainer. Defaults value is `None`.
    client : Dask Client, optional
        Specifies that Dask Client object to use in MLX. If `None`, no parallelization.
    random_state : int, optional
        Random seed, by default 42.

    Return
    ------
    :class:`mlx.PermutationImportance`
    """
    if explainer is None:
        exp = PermutationImportance()
    else:
        if not isinstance(explainer, PermutationImportance):
            raise TypeError(
                "Invalid explainer provided to "
                "init_permutation_importance_explainer: {}".format(type(explainer))
            )
        exp = explainer
    if check_tabular_or_text(est, X_train) == "text":
        raise TypeError(
            "Global feature importance explainers are currently not "
            "supported for text datasets."
        )
    exp.set_config(mode=mode, client=client, random_state=random_state)
    try:
        exp.fit(
            est,
            X_train,
            y=y_train,
            target_names=class_names,
            selected_features=selected_features,
        )
    except Exception as e:
        logger.error(
            f"Unable to construct the PermutationImportance explainer based on the MLX config "
            f"and fit to the train data due to: {e}."
        )
        raise e
    return exp


@deprecated(
    details="Working with AutoML has moved from within ADS to working directly with the AutoMLx library. AutoMLx are preinstalled in conda pack automlx_p38_cpu_v2 and later, and can now be updated independently of ADS. AutoMLx documentation may be found at https://docs.oracle.com/en-us/iaas/tools/automlx/latest/html/multiversion/v23.1.1/index.html. Notebook examples are in Oracle's samples repository: https://github.com/oracle-samples/oci-data-science-ai-samples/tree/master/notebook_examples and a migration tutorial can be found at https://accelerated-data-science.readthedocs.io/en/latest/user_guide/model_training/automl/quick_start.html .",
    raise_error=True,
)
def init_partial_dependence_explainer(
    explainer, est, X_train, y_train, mode, class_names=None, client=None
):
    """
    Initializes a Global feature dependence explainer.

    Supports one and two feature partial dependence plots (PDP) and one feature individual
    conditional expectation plots (ICE). Currently only supported for tabular datasets
    (text is not supported).

    The explainer is initialized to the defaults and can be updated with
    `MLXGlobalExplainer.configure_partial_dependence()`.

    Parameters
    ----------
    explainer : FDExplainer
        If the explainer has previously been initialized, it can be passed in to avoid
        creating a new explainer object. If None, a new `FDExplainer` will be created.
    est : ADSModel
        Model to explain.
    X_train : pandas.DataFrame
        Training dataset.
    y_train : pandas.DataFrame/Series
        Training labels.
    mode : str
        'classification' or 'regression'.
    class_names : list
        List of target names.
    client : Dask Client
        Specifies that Dask Client object to use in MLX. If None, no parallelization.

    Return
    ------
    :class:`mlx.FDExplainer`
    """
    if explainer is None:
        exp = FDExplainer()
    else:
        if not isinstance(explainer, FDExplainer):
            raise TypeError(
                "Invalid explainer provided to "
                "init_partial_dependence_explainer: {}".format(type(explainer))
            )
        exp = explainer
    if check_tabular_or_text(est, X_train) == "text":
        raise TypeError(
            "Global partial dependence explainers are currently not "
            "supported for text datasets."
        )
    exp.set_config(mode=mode, client=client)
    exp.fit(est, X_train, y=y_train, target_names=class_names)
    return exp


@deprecated(
    details="Working with AutoML has moved from within ADS to working directly with the AutoMLx library. AutoMLx are preinstalled in conda pack automlx_p38_cpu_v2 and later, and can now be updated independently of ADS. AutoMLx documentation may be found at https://docs.oracle.com/en-us/iaas/tools/automlx/latest/html/multiversion/v23.1.1/index.html. Notebook examples are in Oracle's samples repository: https://github.com/oracle-samples/oci-data-science-ai-samples/tree/master/notebook_examples and a migration tutorial can be found at https://accelerated-data-science.readthedocs.io/en/latest/user_guide/model_training/automl/quick_start.html .",
    raise_error=True,
)
def init_ale_explainer(
    explainer, est, X_train, y_train, mode, class_names=None, client=None
):
    """
    Initializes a Global Accumulated Local Effects(ALE) Explainer.

    Supports one feature ALE plots. Supported for tabular datasets
    (text is not supported).

    The explainer is initialized to the defaults and can be updated with
    `MLXGlobalExplainer.configure_accumulated_local_effects()`.

    Parameters
    ----------
    explainer : AleExplainer
        If the explainer has previously been initialized, it can be passed in to avoid
        creating a new explainer object. If None, a new AleExplainer will be created.
    est : ADSModel
        Model to explain.
    X_train : pandas.DataFrame
        Training dataset.
    y_train : pandas.DataFrame/Series
        Training labels.
    mode : str
        "classification" or "regression".
    class_names : list, optional
        List of target names. Default value is `None`.
    client : Dask Client, optional
        Specifies that Dask Client object to use in MLX. If `None`, no parallelization.

    Return
    ------
    :class:`mlx.FDExplainer`
    """
    if explainer is None:
        exp = AleExplainer()
    else:
        if not isinstance(explainer, AleExplainer):
            raise TypeError(
                "Invalid explainer provided to "
                "init_partial_dependence_explainer: {}".format(type(explainer))
            )
        exp = explainer
    if check_tabular_or_text(est, X_train) == "text":
        raise TypeError(
            "Global partial dependence explainers are currently not "
            "supported for text datasets."
        )
    exp.set_config(mode=mode, client=client)
    exp.fit(est, X_train, y=y_train, target_names=class_names)
    return exp


@deprecated(
    details="Working with AutoML has moved from within ADS to working directly with the AutoMLx library. AutoMLx are preinstalled in conda pack automlx_p38_cpu_v2 and later, and can now be updated independently of ADS. AutoMLx documentation may be found at https://docs.oracle.com/en-us/iaas/tools/automlx/latest/html/multiversion/v23.1.1/index.html. Notebook examples are in Oracle's samples repository: https://github.com/oracle-samples/oci-data-science-ai-samples/tree/master/notebook_examples and a migration tutorial can be found at https://accelerated-data-science.readthedocs.io/en/latest/user_guide/model_training/automl/quick_start.html .",
    raise_error=True,
)
def init_whatif_explainer(
    explainer,
    est,
    X_test,
    y_test,
    mode,
    class_names=None,
    train=None,
    target_title="target",
    random_state=42,
    **kwargs,
):
    if explainer is None:
        width = kwargs.get("width", 1100)
        exp = WhatIf(mode=mode, random_state=random_state, width=width)
    else:
        if not isinstance(explainer, WhatIf):
            raise TypeError(
                "Invalid explorer provided to "
                "init_explorer: {}".format(type(explainer))
            )
        exp = explainer
    exp_type = check_tabular_or_text(est, X_test)
    if exp_type == "text":
        raise TypeError(
            "WhatIf explainer are currently not "
            "supported for text datasets.".format(type(explainer))
        )
    exp.fit(
        model=est,
        X=X_test,
        y=y_test,
        target_names=class_names,
        train=train,
        target_title=target_title,
    )
    return exp



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

"""
NOTE:

    There's an opportunity here to generate a new feature, credict card numbers are not preditive because they
    don't generalize, however, if the feature is replaced by the type of card that might be predictive.

    - Visa: ^4[0-9]{12}(?:[0-9]{3})?$ All Visa card numbers start with a 4. New cards have 16 digits. Old cards have 13.
    - MasterCard: ^(?:5[1-5][0-9]{2}|222[1-9]|22[3-9][0-9]|2[3-6][0-9]{2}|27[01][0-9]|2720)[0-9]{12}$ MasterCard numbers
        either start with the numbers 51 through 55 or with the numbers 2221 through 2720. All have 16 digits.
    - American Express: ^3[47][0-9]{13}$ American Express card numbers start with 34 or 37 and have 15 digits.
    - Diners Club: ^3(?:0[0-5]|[68][0-9])[0-9]{11}$ Diners Club card numbers begin with 300 through 305, 36 or 38.
        All have 14 digits. There are Diners Club cards that begin with 5 and have 16 digits. These are a joint
        venture between Diners Club and MasterCard, and should be processed like a MasterCard.
    - Discover: ^6(?:011|5[0-9]{2})[0-9]{12}$ Discover card numbers begin with 6011 or 65. All have 16 digits.
    - JCB: ^(?:2131|1800|35\d{3})\d{11}$ JCB cards beginning with 2131 or 1800 have 15 digits.
        JCB cards beginning with 35 have 16 digits.

"""

from __future__ import print_function, absolute_import, division

import re

import pandas as pd

from ads.type_discovery import logger
from ads.type_discovery.abstract_detector import AbstractTypeDiscoveryDetector
from ads.type_discovery.typed_feature import CreditCardTypedFeature


class CreditCardDetector(AbstractTypeDiscoveryDetector):

    _max_sample_size_to_luhn_check = 1000
    _pattern_string = r"""^(?:4[0-9]{12}(?:[0-9]{3})?            # Visa
                             |  (?:5[1-5][0-9]{2}                # MasterCard
                                 | 222[1-9]|22[3-9][0-9]|2[3-6][0-9]{2}|27[01][0-9]|2720)[0-9]{12}
                             |  3[47][0-9]{13}                   # American Express
                             |  3(?:0[0-5]|[68][0-9])[0-9]{11}   # Diners Club
                             |  6(?:011|5[0-9]{2})[0-9]{12}      # Discover
                             |  (?:2131|1800|35\d{3})\d{11}      # JCB
                             |  (5018|5020|5038|5612|5893|6304|6759|6761|6762|6763|0604|6390)\d+$   # Maestro
                             |  ^(5[06789]|6)[0-9]{0,}$          # Maestro
                             |  ^4[0-9]{12}(?:[0-9]{6})?$        #Visa 19 digit
                            )$"""

    def luhn_checksum(self, card_number):
        def digits_of(n):
            return [int(d) for d in str(n)]

        digits = digits_of(card_number)
        odd_digits = digits[-1::-2]
        even_digits = digits[-2::-2]
        checksum = 0
        checksum += sum(odd_digits)
        for d in even_digits:
            checksum += sum(digits_of(d * 2))
        return checksum % 10

    def is_luhn_valid(self, card_number):
        return self.luhn_checksum(card_number) == 0

    def is_credit_card(self, name, values):
        cc = re.compile(CreditCardDetector._pattern_string, re.VERBOSE)
        # since the nulls have been previously filtered we can safely do "all"
        samp = (
            values
            if values.size <= CreditCardDetector._max_sample_size_to_luhn_check
            else values.sample(n=CreditCardDetector._max_sample_size_to_luhn_check)
        )

        if samp.dtype.name in ["float16", "float32", "float64"]:
            if samp.apply(float.is_integer).all():
                samp = samp.fillna(0.0).astype(int)

        if samp.dtype.name in ["int16", "int32", "int64"]:
            samp = samp.astype(str)

        if all([cc.match(str(x)) for x in samp]):
            #
            # iff the pattern matching succeeds do we try the luhn algorithm on a sample
            #
            return all([self.is_luhn_valid(x) for x in samp])

        return False

    def discover(self, name, series):
        candidates = series.loc[~series.isnull()]

        if self.is_credit_card(name, candidates.head(1000)):
            logger.debug("column [{}]/[{}] credit card".format(name, series.dtype))
            return CreditCardTypedFeature.build(name, series)

        return False
